from __future__ import print_function
#
#  Name   :ifigure_app
#
#          main application for a script driven
#          data analysis/visulaization tool
#
#  Author :
#         Syun'ichi Shiraiwa
#  E-mail :
#         shiraiwa@psfc.mit.edu
#
#

import time
import weakref
from ifigure.widgets.undo_redo_history import GlobalHistory, UndoRedoHistory
from collections import deque
from ifigure.utils.wx3to4 import (PyDeadObjectError,
                                  menu_Append,
                                  menu_AppendSubMenu)
import ifigure.server
import ifigure.utils.cbook as cbook
from ifigure.widgets.appearance_config import AppearanceConfig
from ifigure.widgets.advanced_config import AdvancedConfig
from ifigure.utils.helper_app import HelperApp
from ifigure.ifigure_config import iFigureConfig
from ifigure.utils.postoffice import PostOffice
from ifigure.mto.py_file import PyText
from ifigure.mto.py_code import PyModel
from ifigure.mto.fig_obj import FigObj
from ifigure.mto.fig_axes import FigAxes
from ifigure.mto.fig_page import FigPage
from ifigure.mto.fig_book import FigBook
from ifigure.mto.project_top import ProjectTop
from ifigure.mto.treelink import TreeLink
from ifigure.mto.treedict import TopTreeDict
import ifigure.widgets.dialog as dialog
from ifigure.widgets.book_viewer import BookViewerFrame, BookViewer
from ifigure.widgets.redirect_output import RedirectOutput
from ifigure.widgets.statusbar import StatusBar
from ifigure.widgets.tipwindow import Tipwindow
from ifigure.widgets.logwindow import Logwindow
from ifigure.widgets.command_history import CommandHistory
from ifigure.widgets.consol import Consol
from ifigure.widgets.script_editor import ScriptEditor
from ifigure.widgets.panel_checkbox import PanelCheckbox
from ifigure.widgets.simple_shell import SimpleShell, FakeSimpleShell
from ifigure.widgets.proj_tree_viewer_aui import ProjTreeViewer
from ifigure.widgets.property_editor import property_editor
from ifigure.widgets.canvas.ifigure_canvas import ifigure_canvas
import ifigure.events
from ifigure.utils.mp_tarzip import MPTarzip
import ifigure.utils.debug as debug
import wx
#import wx.py, wx.lib
#

import sys
import shutil
import os
import tarfile
import collections
import logging
import threading
import ifigure.utils.pickle_wrapper as pickle

from numpy import arange, sin, pi
import ifigure

###
# application global
###
ifigure._cursor_config = {}  # cursor configuration
### populated in load_pref
ifigure._visual_config = {}
#                         {'1dcolor1': 'red',
#                          '1dcolor2': 'blue',
#                          '1dalpha' : 0.5,
#                          '1dthick' : '1.0',
#                          '2dalpha' : 0.7,
#                          '2dcolor' : 'grey',
#                          'format' : '{:.2e}'}
ifigure._cursor_book = None  # specail book for slice cursor

redirect_std = False
use_console = False
#
#  debug setting
#
dprint1, dprint2, dprint3 = debug.init_dprints('iFigureApp')

#from ifigure.mdsplus.mdsscope import MDSScope


try:
    from wx import glcanvas
    haveGLCanvas = True
except ImportError:
    haveGLCanvas = False

try:
    # The Python OpenGL package can be found at
    # http://PyOpenGL.sourceforge.net/
    from OpenGL.GL import *
    #from OpenGL.GLUT import *
    haveOpenGL = True
except ImportError:
    haveOpenGL = False


class WindowList(list):
    def get_list(self):
        ret = [item() for item in self if item() is not None]
        self = [item for item in self if item() is not None]
        return ret

    def add_item(self, item):
        self.append(weakref.ref(item))

    def remove_item(self, item):
        t = None
        for k in self:
            if k() is item:
                t = k
        if t is not None:
            self.remove(t)

    def get_next(self, current):
        ret = self._get_valid_item()
        self._validate_ref()
        id = ret.index(current) + 1
        if id >= len(self):
            id = 0
        return self[id]()

    def get_prev(self, current):
        ret = self._get_valid_item()
        self._validate_ref()
        id = ret.index(current) - 1
        if id < 0:
            id = len(self) - 1
        return self[id]()

    def _get_valid_item(self):
        self._validate_ref()
        ret = [item() for item in self]
        return ret

    def _validate_ref(self):
        # check for wx4
        ret = [item for item in self if item() is not None]
        self[:] = [item for item in ret if item()]
        # check for wx3
        ret = [item for item in self if item() is not None]
        self[:] = [item for item in ret if isinstance(item(), wx.Window)]


class CanvasPanel(wx.Panel):
    pass


class EditorPanel(wx.Panel):
    pass


ID_DETACH_EDITOR = wx.NewIdRef(count=1)
ID_SAVEDOC = wx.NewIdRef(count=1)
ID_SAVEASDOC = wx.NewIdRef(count=1)

RECENT_FILE = deque([''] * 10, 10)


class ifigure_app(BookViewerFrame):
    def __init__(self, parent, title, noPyShell=False, hide=False,
                 launcher_file=None):

        self.windowlist = WindowList()
        self.appearanceconfig = AppearanceConfig()

        mb = wx.MenuBar()
        wx.MenuBar.MacSetCommonMenuBar(mb)

        super(ifigure_app, self).__init__(parent,
                                          title=title, size=(10, 10),
                                          style=wx.MINIMIZE_BOX | wx.MAXIMIZE_BOX |
                                          wx.RESIZE_BORDER | wx.SYSTEM_MENU |
                                          wx.CAPTION | wx.CLOSE_BOX |
                                          wx.CLIP_CHILDREN)
        self._launcher_file = launcher_file
        self._ID_RECENT = -1
        logging.basicConfig(level=logging.DEBUG)
        self.stdfiles = (sys.stdout, sys.stderr)
        self._text_clip = ''
        self.timer = None
        self.remote_lock = threading.Lock()
        self.viewers = [self]  # list of fig viewers
        self._aviewer = self

        self.ipage = -1
        self.book = self.make_empty_project()
        self.load_pref()
        self.InitUI(parent, title, noPyShell=noPyShell)

        self.BindTreeDictEvents()
        self.BindPVCVEvents()
#       self.Bind(ifigure.events.TD_EVT_NEWHISTORY, self.onNewHistory)

        self.shell.set_proj(self.proj)

        self.set_filename_2_window_title()
        self.logw = Logwindow(self, wx.ID_ANY, 'log')
#       self.logw.redirector = RedirectOutput(sys.stdout)
        self.tipw = Tipwindow(self, wx.ID_ANY, 'Command help')
        self.shell.set_tipw(self.tipw)
        self.hdf_export_window = None

        if use_console:
            self.redirector = RedirectOutput(self.proj_tree_viewer.consol.log,
                                             self.proj_tree_viewer.consol.log)
#           sys.stdout = redirector
#           sys.stderr = redirector
        else:
            self.redirector = RedirectOutput(sys.stdout, sys.stderr)
        if redirect_std:
            self.redirector.turn_force_on()        # doesn't do anything if not redirect_std
        else:
            self.redirector.turn_on()  # redirect anyway
        self.logw.set_redirector(self.redirector)

        # these are preference components
        self.po = PostOffice().set_parent(self)
        self.config = iFigureConfig()
        self.helper = HelperApp()
        self.aconfig = AdvancedConfig()
        self.aconfig.add_user_path()

        self.SetMenuBar(self.menuBar)

        self.Layout()

        if not hide:
            wx.CallAfter(self.Show, True)
            wx.CallAfter(self.Raise)
            wx.CallAfter(
                self.proj_tree_viewer.get_shellvar_viewer().update, self.shell)
        wx.CallAfter(self.adjust_initial_position)

    def adjust_initial_position(self):
        dx, dy = wx.GetDisplaySize()
        msx, msy = self.GetSize()

        if msy > dy:
            msy = dy - 150
        if msx > dx:
            msx = dx - 250
        self.SetSize((msx, msy))
        self.Layout()
        h, w = self.GetPosition()
        if h < 0 or w < 0:
            self.SetSize((msx - (100 - h), msy - (100 - h)))
            self.SetPosition((100, 100))

        wx.CallAfter(self.CentreOnScreen)

    @property
    def aviewer(self):
        return self._aviewer

    @aviewer.setter
    def aviewer(self, value):
        self._aviewer = value
        if value is not None:
            if not value.isinteractivetarget:
                return
        from ifigure.interactive import set_aviewer
        set_aviewer(self._aviewer)

    def get_components(self):
        return [self.config, self.appearanceconfig,
                self.po, self.helper, self.aconfig]

    def load_pref(self):
        from ifigure.utils.setting_parser import iFigureSettingParser as SP
        p = SP()
        names = ['cursor_config', 'visual_config']
        for name in names:
            var = p.read_setting('pref.' + name)
            d = getattr(ifigure, '_' + name)
            for key in var:
                d[key] = var[key]

    def save_pref(self):
        from ifigure.utils.setting_parser import iFigureSettingParser as SP
        p = SP()
        names = ['cursor_config', 'visual_config']
        for name in names:
            d = getattr(ifigure, '_' + name)
            p.write_setting('pref.' + name, d)

    def InitUI(self, parent, title, noPyShell=False):
        # A Statusbar in the bottom of the window
        self.sb = StatusBar(self)
        self.SetStatusBar(self.sb)

        # define splitter panel tree

        # valid panel p1, p22, p121, p122
        self.gui_tree = PanelCheckbox(self, wx.VERTICAL)
        p1, p2 = self.gui_tree.add_splitter('v', 'h')
        p21, p22 = p2.add_splitter('h', 'h')
        p211, p212 = p21.add_splitter('v', 'h')
        # make all panels
        self.proj_tree_viewer = p1.add_panel(ProjTreeViewer,
                                             "Project", "Project", 0, 'l',
                                             1, wx.ALL | wx.EXPAND, 0)
        if noPyShell:
            shell = FakeSimpleShell
        else:
            shell = SimpleShell
        self.shell = p22.add_panel(shell,
                                   "Shell", "Shell", 1, 'b',
                                   1, wx.ALL | wx.EXPAND, 0)

        self.panel2 = p211.add_panel(EditorPanel,
                                     "Editor", "Editor", 2, False,
                                     1, wx.ALL | wx.EXPAND, 0)

        self.panel2.SetSizer(wx.BoxSizer(wx.VERTICAL))
        self.make_attached_script_editor()
        self.panel1 = p212.add_panel(CanvasPanel,
                                     "Figure", "Figure", 3)
        p212.set_primary(self.panel1)

        self.property_editor = p212.add_panel(property_editor,
                                              "Property", "Property",
                                              4, 'r', 0, wx.ALL | wx.EXPAND, 0)
        # self.gui_tree.hide_toggle_menu(self.panel1)
        # self.gui_tree.hide_toggle_menu(self.property_editor)
        ch = self.proj_tree_viewer.get_command_history()
        self.shell.set_command_history(ch)
        # self.script_editor.Hide()
        self.canvas = None
        self._rebuild_ifigure_canvas()
        self._link_canvas_property_editor()
        self.gui_tree.primary_client(self.canvas)

        # File Menu
        newmenu = wx.Menu()
        menu_AppendSubMenu(self.filemenu, wx.ID_ANY, 'New', newmenu)
        #menu_Append(self.filemenu, wx.ID_ANY, 'New', newmenu)
        self.add_menu(newmenu, wx.ID_ANY,
                      "piScope", "Start new piScope application",
                      self.onNewApp)
        self.add_menu(newmenu, wx.ID_ANY,
                      "Project", "Create new project",
                      self.onNew)
        self.add_menu(newmenu, wx.ID_ANY,
                      "Figure", "Create new book and open it in a new viewer",
                      self.onNewBook)
        self.add_menu(newmenu, wx.ID_ANY,
                      "Script", "Create new script in project",
                      self.onNewScript)
        self.add_menu(newmenu, wx.ID_ANY,
                      "Text", "Create new text in project",
                      self.onNewFile)
        self.add_menu(newmenu, wx.ID_ANY,
                      "non-project Text", "Create new untitled text (file is not stored in project)",
                      self.onNewDoc)
        openmenu = wx.Menu()
        menu_AppendSubMenu(self.filemenu, wx.ID_ANY, 'Open', openmenu)
        #menu_Append(self.filemenu, wx.ID_ANY, 'Open', openmenu)
        self.add_menu(openmenu, wx.ID_ANY,
                      "Project...", "Open an existing project",
                      self.onOpen)
        self.add_menu(openmenu, wx.ID_ANY,
                      "Project in new piScope...", "Open an existing project in new piscope",
                      self.onOpenInNewpiScope)
        self.add_menu(openmenu, wx.ID_OPEN,
                      "Book ...",
                      "Import book file (.bfz). Current book is deleted from project",
                      self.onLoadBook)
        self.add_menu(openmenu, wx.ID_ANY,
                      "Book in new window...",
                      "Import Book file (.bfz), New book data will be added to project",
                      self.onLoadBookNew)
        self.add_menu(openmenu, wx.ID_ANY,
                      "Script...", "Open Script",
                      self.onOpenScript)
        self.add_menu(openmenu, wx.ID_ANY,
                      "File...", "Open File",
                      self.onOpenFile)
        self._recentmenu = wx.Menu()
        item = self.filemenu.AppendSubMenu(self._recentmenu, "Open Recent")
        self._ID_RECENT = item.GetId()
        # menu_AppendSubMenu(self.filemenu, ID_RECENT,
        #            "Open Recent", self._recentmenu)
        # m = self.filemenu.AppendSubMenu(self._recentmenu,
        #                                "Open Recent")
        #globals()['ID_RECENT'] = m.GetId()

        self.filemenu.AppendSeparator()
        self.append_save_project_menu(self.filemenu)
        self.add_menu(self.filemenu, ID_SAVEDOC,
                      "Save Document", "Save current editor document",
                      self.onSaveDoc)
        self.add_menu(self.filemenu, ID_SAVEASDOC,
                      "Save Document As...",
                      "Save current editor document in a different file",
                      self.onSaveAsDoc)
#       self.add_menu(self.filemenu, wx.ID_ANY,
#                     "Close Project", "Close Project",
#                     self.onCloseProject)
        self.add_menu(self.filemenu, wx.ID_ANY,
                      "Preference...", "Preference",
                      self.onAppPreference)
        self.filemenu.AppendSeparator()
        self.export_book_menu = self.add_menu(self.filemenu,
                                              BookViewerFrame.ID_EXPORTBOOK,
                                              "Export Book", "Export Book",
                                              self.onExportBook)
#       self.export_book_menu.Enable(True)
        self.add_menu(self.filemenu,
                      BookViewerFrame.ID_EXPORTBOOK_AS,
                      "Export Book As...", "Export Book",
                      self.onExportBookAs)
        self.add_saveimage_menu(self.filemenu)
        self.filemenu.AppendSeparator()
        self._figure_mnis = [self.add_menu(self.filemenu, wx.ID_ANY,
                                           "Close Figure", "Close figure",
                                           self.onCloseFigure), ]
#                                 self.add_menu(self.filemenu, wx.ID_ANY,
#                                              "Close Figure + Delete Book",
#                                    "Close figure and delete book data",
#                                        self.onCloseFigureDeleteBook)]
        self.filemenu.AppendSeparator()
        self.add_quitemenu(self.filemenu)
#       self.add_menu(self.filemenu, wx.ID_ANY,
#                     "Quit piScope"," Terminate the program",
#                     self.onQuit)
        self.Bind(wx.EVT_CLOSE, self.onAppWindowClose)
        # plot menu
#       self.plotmenu = wx.Menu()
#       self.menuBar.Append(self.plotmenu,"Plot")
        self.add_std_plotmenu(self.plotmenu)
        # help menu
        self.append_help_menu()

        # Edit Menu
        self.append_undoredo_menu(self.editmenu)
        self.editmenu.AppendSeparator()

        self.add_cutpaste_menu(self.editmenu)

        panelmenu = wx.Menu()
        menu_AppendSubMenu(self.viewmenu, wx.ID_ANY, 'Panels', panelmenu)
        #menu_Append(self.viewmenu, wx.ID_ANY, 'Panels', panelmenu)
        self.gui_tree.append_menu(panelmenu)
        self.viewmenu.AppendSeparator()
        self.gui_tree.update_check()
        self.gui_tree.bind_handler(self)

        self.gui_tree.set_splitters()

        self.editmenu.AppendSeparator()
        self.add_bookmenus(self.editmenu, self.viewmenu)
        self.viewmenu.AppendSeparator()
        x = self.add_menu(self.viewmenu, wx.ID_ANY,
                          "Detach Figure", " Detach figure and show it in a separate window",
                          self.onDetachFigure)
        self._figure_mnis.append(x)
        self.add_menu(self.viewmenu, ID_DETACH_EDITOR,
                      "Detach Editor", " Detach editor in a separate window",
                      self.onDetachEditor)

        self.append_std_viewmenu2(self.viewmenu)
#       self.editmenu.AppendSeparator()
#       self.append_undoredo_menu(self.editmenu)
        self.editmenu.AppendSeparator()
        self.add_menu(self.editmenu, wx.ID_ANY,
                      "Clear history", "Clear command history panel",
                      self.onClearCommandHistory)
#       self.add_menu(self.editmenu, wx.ID_ANY,
#                     "MP","testing mp",
#                     self.onMP)
#       self.add_menu(self.editmenu, wx.ID_ANY,
#                     "MP End","testing mp",
#                     self.onMPEnd)

        self.helpmenu.AppendSeparator()
        self.add_menu(self.helpmenu, wx.ID_HELP,
                      "About...", "About this program",
                      self.onAbout)

#       self.SetMenuBar(self.menuBar)
        # Adding the MenuBar to the Frame content.
        # aTable = wx.AcceleratorTable([(wx.ACCEL_CTRL,  ord('Z'), wx.ID_UNDO),
        #                (wx.ACCEL_CTRL|wx.ACCEL_SHIFT, ord('Z'), wx.ID_REDO),
        #                ])
        # self.SetAcceleratorTable(aTable)

#       self.Layout()

        self.panel2.Hide()
        self.gui_tree.hide_toggle_menu(self.panel2)
        self.gui_tree.rebuild_menu()

        self.SetSize([800, 500])
        self.gui_tree.update_check()
        self.gui_tree.set_splitters()
        self.Layout()
        size = self.GetSize()
        self.gui_tree.set_sashposition([200, size[1] - 100, 300])
        self.property_editor.set_sizehint()

  #      if os.path.exists(geom_file):
        from ifigure.ifigure_config import geom_file
        try:
            val = self.read_geom_file()
            if "editor_detached" in val:
                #               print 'editor_detached', val["editor_detached"]
                self.script_editor._first_open_style = val["editor_detached"]
        except BaseException:
            self.book.set_open(True)
        else:
            val["sh"][1] = False  # this is to hide script editor

            # safe guard in case size is zero...
            val["size"][0] = max((val["size"][0], 300))
            val["size"][1] = max((val["size"][1], 300))
            self.SetSize(val["size"])
            self.gui_tree.set_showhide(val["sh"])
            self.gui_tree.update_check()
            self.gui_tree.set_splitters()

            self.Layout()
            self.gui_tree.set_sashposition(val["pos"])
#           self.property_editor.set_sizehint()

            if not val["sh"][2]:
                #               self.close_figurebook()
                self.onCloseFigureDeleteBook()
            else:
                self.aviewer = self
                self.book.set_open(True)
            if 'RECENT_FILE' in val:
                for item in val['RECENT_FILE']:
                    if not item in RECENT_FILE and os.path.exists(item):
                        RECENT_FILE.append(item)

        self.proj_tree_viewer.update_widget()
        self.set_accelerator_table()

    def onUpdateUI(self, evt):
        if evt.GetId() == ID_DETACH_EDITOR:
            if (self.script_editor.GetTopLevelParent() == self and
                    self.gui_tree.get_toggle(self.panel2)):
                evt.Enable(True)
            else:
                evt.Enable(False)
        elif evt.GetId() == self._ID_RECENT:
            m = self._recentmenu
            for item in m.GetMenuItems():
                m.DestroyItem(item)
            evt.Enable(True)
            mm = []
            self.read_recent_files()
            for item in RECENT_FILE:
                if item == '':
                    continue

                def dummy(evt, file=item):
                    print("recent menu", file)
                    self.onOpen(None, path=file)
                    evt.Skip()
                mm.append((  # os.path.basename(item),
                    item,
                          'Open ' + item,
                          dummy))
            if len(mm) > 0:
                for a, b, c in mm:
                    mmi = self.add_menu(m, wx.ID_ANY, a, b, c)
                m.AppendSeparator()
                mmi = self.add_menu(m, wx.ID_ANY, 'Reset Menu',
                                    'Reset recent file menu', self.onResetRecent)
            else:
                evt.Enable(False)
        elif evt.GetId() == wx.ID_UNDO:
            fc = self.FindFocus()
            if isinstance(fc, wx.stc.StyledTextCtrl):
                evt.SetText('Undo text edit')
            else:
                self.history.update_menu_item()
        elif evt.GetId() == wx.ID_REDO:
            fc = self.FindFocus()
            if isinstance(fc, wx.stc.StyledTextCtrl):
                evt.SetText('Redo text edit')
            else:
                self.history.update_menu_item()
        elif evt.GetId() == ID_SAVEDOC:
            fc = self.FindFocus()
            if fc in self.script_editor.page_list:
                a, b = self.script_editor.check_if_need_save(fc)
                if a:
                    evt.Enable(True)
                else:
                    evt.Enable(False)
            else:
                evt.Enable(False)
        elif evt.GetId() == ID_SAVEASDOC:
            fc = self.FindFocus()
            if fc in self.script_editor.page_list:
                evt.Enable(True)
            else:
                evt.Enable(False)
        else:
            return super(ifigure_app, self).onUpdateUI(evt)

    def BindPVCVEvents(self):
        self.Bind(ifigure.events.PV_EVT_DrawRequest,
                  self.onPV_DrawRequest)
        self.Bind(ifigure.events.PV_EVT_AddFigobj,
                  self.onPV_AddFigobj)
        self.Bind(ifigure.events.PV_EVT_DeleteFigobj,
                  self.onPV_DeleteFigobj)
        self.Bind(ifigure.events.CV_EVT_CanvasSelected,
                  self.onCV_CanvasSelected)
        self.Bind(ifigure.events.REMOTE_COMMAND,
                  self.onRemoteCommand)
        self.Bind(ifigure.events.TD_EVT_UNDO,
                  self.onTD_EvtUndo)
        self.Bind(ifigure.events.TD_EVT_REDO,
                  self.onTD_EvtRedo)
        self.Bind(ifigure.widgets.simple_shell.EVT_SHELL_ENTER,
                  self.onShellEnter)
        self.Bind(ifigure.events.TD_EVT,
                  self.onTD_Evt)
        self.Bind(ifigure.events.TD_EVT_FILESYSTEMCHANGED,
                  self.onFileSystemChanged)
        self.Bind(ifigure.events.TD_EVT_PAGESHOWN,
                  self.onTD_PageShown)
        self.Bind(ifigure.events.TD_EVT_CLOSEBOOKREQUEST,
                  self.onTD_CloseBookRequest)

    def onFileSystemChanged(self, evt=None):
        self.script_editor.onFileSystemChanged(evt)
    #
    #  File Menu
    #

    def onResetRecent(self, evt):
        for i in range(len(RECENT_FILE)):
            RECENT_FILE.append('')
        self.write_recent_files()

    def onExportBook(self, evt):
        self.onSaveBook(evt)

    def onExportBookAs(self, evt):
        self.onSaveBookAs(evt)

    def onNewBook(self, evt, veiwer=None):
        from ifigure.widgets.book_viewer import BookViewer
        super(ifigure_app, self).onNewBook(evt,
                                           viewer=BookViewer, proj=self.proj)

    def onLoadBook(self, evt, mode=0, proj=None):
        if self.book is None:
            bk = super(ifigure_app, self).onLoadBook(evt, mode=1,
                                                     proj=self.proj)
            if bk is not None:
                self.open_book_in_appwindow(bk)
        else:
            super(ifigure_app, self).onLoadBook(evt)

    def onLoadBookNew(self, evt, viewer=None, proj=None):
        from ifigure.widgets.book_viewer import BookViewer
        super(ifigure_app, self).onLoadBookNew(evt,
                                               viewer=BookViewer,
                                               proj=self.proj)

    def onOpen(self, e=None, path=None):
        from ifigure.utils.mp_tarzip import local_lc
        if not local_lc.acquire(False):
            ret = dialog.message(
                self, 'Save job is running.', 'Please wait', 0)
            return
        local_lc.release()

        if path is None:
            path = dialog.read(parent=self,
                               message="Select project (.pfz) to open",
                               wildcard='*.pfz',
                               defaultdir=os.getcwd())
        if path != '':
            call_close = (self.proj is not None)
            if not self.open_file(path, call_close=call_close):
                return
            self.deffered_force_layout()
            self.set_proj_saved(True)

    def get_file_helper_command(self, path):
        from ifigure.utils.setting_parser import iFigureSettingParser as SP
        p = SP().set_rule('file',
                          {'name': '', 'ext': '', 'action': '',
                           'action_txt': '', 'use': False})
        var = p.read_setting('pref.file_helper')
        import os
        path = os.path.expanduser(path)
        basepath = os.path.basename(path)
        import re
        for x in var['file']:
            try:
                rule = x['ext']
                p = re.compile(rule)
            except BaseException:
                print('compile error of regular expression: ' + rule)
            if p.match(basepath) is not None:
                if not x['use']:
                    continue
                command = x['action']
                command = command.replace('{1}', '"' + path + '"')
                command = command.replace('{top}', 'proj')
                return command
        return ''

    def onOpenWithHelperCommand(self, e=None, path=None, hide_main=False):
        command = self.get_file_helper_command(path)
        if command != '':
            #            self.shell.lvar['command'] = command
            wx.CallAfter(self.shell.Execute, command)
            wx.CallAfter(self.proj_tree_viewer.update_widget)
            if hide_main:
                wx.CallAfter(self.goto_no_mainwindow)

    def onOpenInNewpiScope(self, e):
        path = dialog.read(parent=self,
                           message="Select project (.pfz) to open in new piScope application",
                           wildcard='*.pfz')
        if path != '':
            import ifigure
            import subprocess
            import shlex
            import ifigure.widgets.canvas.ifigure_canvas
            options = ' '
            if not ifigure.widgets.canvas.ifigure_canvas.turn_on_gl:
                options = options + '-g '
            if redirect_std:
                options = options + '-d '

            script = os.path.join(os.path.dirname(ifigure.__file__),
                                  "__main__.py ")
            command = sys.executable + ' ' + script + options + path
            if os.altsep is not None:
                command = command.replace(os.sep, os.altsep)

            self.redirector.turn_off()
            p = subprocess.Popen(shlex.split(command))
#                           shell = True)
#                           stdout=subprocess.PIPE)
            self.redirector.turn_on()

    def open_file(self, file, call_close=False):
        current_file = self.proj.getvar('filename')
        if current_file == file:
            return

        tmp_top = TopTreeDict()
        tmp_top.set_app(self)
        proj = tmp_top.LoadFromFile(file, sb=self.sb)
        if proj is None:
            print("file load error")
            return False
        if call_close:
            self.onCloseProject()

        self.read_recent_files()
        if not file in RECENT_FILE:
            RECENT_FILE.append(file)
            self.write_recent_files()

        self.Freeze()

        self._set_proj(proj)

        #        if (proj.pbook is not None and
        #            proj.pbook.num_page() != 0):
        #           self.book = proj.pbook
        #           self.ipage=0
        #           self.show_page(0)
        #           self.proj_tree_viewer.update_widget()

        # open previously opend files
        from ifigure.mto.treedict import str_td
        wdir = self.proj.getvar('wdir')
        dot_file = os.path.join(wdir, '.opened_file')
        file_count = 0
        if os.path.exists(dot_file):
            fid = open(dot_file, 'r')
            for name in fid.readlines():
                file = name.strip('\n').split('\t')[0]
                file = os.path.join(wdir, file)
                # note: if file is absolute path, wdir is ignored
                if not os.path.exists(file):
                    continue
                if len(name.strip('\n').split('\t')) == 2:
                    file = str_td(file)
                    file.td = name.strip('\n').split('\t')[1]
                ifigure.events.SendEditFileEvent(self.proj,
                                                 w=self,
                                                 file=file)
                file_count = file_count + 1
            fid.close()
        if file_count == 0:
            # close editor if it is shown
            if not self.isEditorAttached():
                self.attach_editor(open_editor=True)
            self.gui_tree.set_showhide([False], self.panel2)
            self.gui_tree.show_toggle_menu(self.panel2)
            self.gui_tree.rebuild_menu()
            self.gui_tree.update_check()
            self.gui_tree.set_splitters()

        dot_file = os.path.join(wdir, '.command_history')
        if os.path.exists(dot_file):
            self.proj_tree_viewer.ch.loadfromfile(dot_file)

        ###
        self.load_gui_setting()
#        self._force_layout()

        show_hg_message = False
        if (self.proj is not None and wx.GetApp().IsMainLoopRunning()):
            self.perform_package_update_check()
        else:
            wx.CallAfter(self.perform_package_update_check)
        # load gui setting
        self.proj_tree_viewer.update_widget()
#        self.update_save_project_menu(True)
        self.Thaw()
        self.sb.SetStatusText("", 0)
        self.sb.Refresh()

        return True

    def perform_package_update_check(self):
        dprint1('checking HG repo for updates')
        from ifigure.mto.hg_support import hg_check_all_incoming_outgoing
        ret, ret2, ret3 = hg_check_all_incoming_outgoing(self.proj)
        if len(ret) > 0 or len(ret2) > 0:
            wx.CallAfter(self.show_package_update_notice,
                         len(ret), len(ret2), len(ret3))

    def show_package_update_notice(self, l, l2, l3):
        dlg = wx.MessageDialog(self,
                               str(l) + " packages are updated in repository.\n" +
                               str(l2) + " local packages are newer than repository.\n" +
                               str(l3) + " packages may need merge (multiple heads).\n\nConsider Package Update from project pull down menu.",
                               'Package check notice',
                               wx.OK | wx.STAY_ON_TOP | wx.ICON_INFORMATION)
        ret = dlg.ShowModal()
        dlg.Destroy()

    def make_empty_project(self):
        '''
        make some default empty project
        '''
        self._set_proj(ProjectTop())
        # return None
        book = self.proj.onAddBook()
        #book._keep_data_in_tree = True
        ipage = book.add_page()
        self.ipage = ipage
        f_page = book.get_page(ipage)
        f_page.add_axes()
        return book

    def open_editor_panel(self):
        if (not self.gui_tree.get_toggle(self.panel2) and
                self.isEditorAttached()):
            self.gui_tree.set_showhide([True], self.panel2)
            self.gui_tree.show_toggle_menu(self.panel2)
            self.gui_tree.rebuild_menu()
            self.gui_tree.update_check()
            self.gui_tree.set_splitters()

    def onOpenFile(self, e=None):
        path = dialog.read(parent=self, message="Select file to open",
                           wildcard='Any|*|py(*.py)|*.py')
        if path != '':
            command = self.get_file_helper_command(path)
            if command != '':
                wx.CallAfter(self.shell.Execute, command)
                wx.CallAfter(self.proj_tree_viewer.update_widget)
            else:
                self.script_editor.OpenFile(path)
                self.open_editor_panel()
#           self.gui_tree.set_splitters()
                self.Layout()

    def onOpenScript(self, e=None):
        path = dialog.read(parent=self, message="Select script to open",
                           wildcard='py(*.py)|*.py')
        if path != '':
            dlg = wx.MessageDialog(None,
                                   'Do you want to create Script in Tree?',
                                   'Import script',
                                   wx.YES_NO)
            ret = dlg.ShowModal()
            dlg.Destroy()
            if ret == wx.ID_YES:
                newname = str(os.path.basename(path).split('.')[0])
                script = self.proj.add_script(newname)
                script.onImportScriptFile(file=path)
                self.proj_tree_viewer.update_widget()
                if e is not None:
                    script.onEditScript(e)
            else:
                self.script_editor.OpenFile(path)
                self.open_editor_panel()
#           self.gui_tree.set_splitters()
                self.Layout()

    def onNew(self, e=None, use_dialog=True):
        if self.proj is not None:
            if use_dialog:
                dlg = wx.MessageDialog(None,
                                       'Do you want to close it?',
                                       'Project is open',
                                       wx.OK | wx.CANCEL)
                ret = dlg.ShowModal()
                dlg.Destroy()
            else:
                ret = wx.ID_OK
            if ret == wx.ID_OK:
                self.onCloseProject(e)
            else:
                return
        book = self.make_empty_project()
#        self.open_book_in_appwindow(book, ipage=0)

        self.proj_tree_viewer.update_widget()
        self.proj_tree_viewer.ch.clear_text()
        self._force_layout()
        self.set_filename_2_window_title()
#        self.update_save_project_menu(False)
        self.set_proj_saved(False)

    def onNewApp(self, e):
        import ifigure
        import subprocess
        import shlex
        import ifigure.widgets.canvas.ifigure_canvas
        options = ' '
        if not ifigure.widgets.canvas.ifigure_canvas.turn_on_gl:
            options = options + '-g '
        if redirect_std:
            options = options + '-d '

        script = os.path.join(os.path.dirname(ifigure.__file__),
                              "__main__.py ")
        command = sys.executable + ' ' + script + options

        if os.altsep is not None:
            command = command.replace(os.sep, os.altsep)

        self.redirector.turn_off()

        p = subprocess.Popen(shlex.split(command))
#                           shell = True)
#                           stdout=subprocess.PIPE)
        self.redirector.turn_on()

    def isEditorAttached(self):
        return self.script_editor.GetTopLevelParent() == self

    def onNewFile(self, e=None):
        obj = self.proj.onAddNewText()
        self.proj_tree_viewer.update_widget()

    def onNewScript(self, e=None):
        obj = self.proj.onAddNewScript()
        self.proj_tree_viewer.update_widget()

    def onNewDoc(self, e=None):
        #        path = obj.path2fullpath()
        if not self.gui_tree.get_toggle(
                self.panel2) and self.isEditorAttached():
            #             self.gui_tree.toggle_panel(self.panel2, True)
            self.gui_tree.set_showhide([True], self.panel2)
            self.gui_tree.show_toggle_menu(self.panel2)
            self.gui_tree.rebuild_menu()
            self.gui_tree.update_check()
            self.gui_tree.set_splitters()

        self.script_editor.NewFile()

    def save_gui_setting(self):
        '''
        save gui preference (if any)  at save time
        '''
        if self.canvas.get_figure() is not None:
            a = self.canvas.get_figure().figobj.get_full_path()
        else:
            a = ''
        data = {'open page': a}

        path = self.proj.getvar("wdir")
        fpath = os.path.join(path, '.gui_setting')
        fid = open(fpath, 'wb')
        pickle.dump(data, fid)
        fid.close()

        data = [{'module': v.__module__,
                 'class': v.__class__.__name__,
                 'path': v.book.get_full_path(),
                 'ipage': v.ipage,
                 'size': v.GetSize(),
                 'pos': v.GetPosition(),
                 'isPropShown': v.isPropShown()}
                for v in self.viewers]
        fpath = os.path.join(path, '.opened_books')
        fid = open(fpath, 'wb')
        pickle.dump(data, fid)
        fid.close()

    def load_gui_setting(self):
        '''
        load gui preference at save time

        gui_setting : currently not used
        opend_books : information of previously opend book and page
        '''
        path = self.proj.getvar("wdir")
        fpath = os.path.join(path, '.gui_setting')
        if os.path.exists(fpath):
            fid = open(fpath, 'rb')
            gui_setting_data = pickle.load(fid)
            fid.close()

        fpath = os.path.join(path, '.opened_books')
        if os.path.exists(fpath):
            fid = open(fpath, 'rb')
            opend_book_data = pickle.load(fid)
            fid.close()
            flag = False
            for v in opend_book_data:
                if isinstance(v, tuple):
                    v = {'module': v[0],
                         'class': v[1],
                         'path': v[2],
                         'ipage': v[3]}
                try:
                    mod = __import__(v['module'])
                    mod = sys.modules[v['module']]
                    cls = getattr(mod, v['class'])
                    book = self.proj.find_by_full_path(v['path'])
#               print cls, self.__class__
                    kargs = {key: v[key] for key in ('ipage', 'size', 'pos', 'isPropShown')
                             if key in v}

                    if cls == self.__class__:
                        self.open_book_in_appwindow(book, ipage=v['ipage'])
                        flag = True
                    else:
                        #                      kargs['size'] = (200, 400)
                        ifigure.events.SendOpenBookEvent(book,
                                                         w=self,
                                                         viewer=cls,
                                                         **kargs)
                except BaseException:
                    import traceback
                    traceback.print_exc()

            if not flag:
                self._go_closed_figure_mode()

    def onSave(self, e):
        from ifigure.utils.mp_tarzip import local_lc
        if not local_lc.acquire(False):
            ret = dialog.message(
                self, 'Previous save job is still running.\n If this is an error, choose unlock. ', 'Please wait (locked)', 2,
                labels=["OK", "Unlock"])
            if ret == 'cancel':
                local_lc.release()
            return

        path = self.proj.getvar("filename")
        if path is None:
            local_lc.release()
            self.onSaveAs(e)
        else:
            self.save_gui_setting()
            self.script_editor.SaveAll()
            self.proj_tree_viewer.ch.savetofile(filename=os.path.join(
                self.proj.getvar("wdir"), '.command_history'))

            self.proj.SaveToFile(path)
        self.set_filename_2_window_title()
        self.set_proj_saved(True)

    def onSaveDoc(self, e):
        ipage = self.script_editor.nb.GetSelection()
        self.script_editor.SaveFile(saveas=False, ipage=ipage)

    def onSaveAsDoc(self, e):
        ipage = self.script_editor.nb.GetSelection()
        self.script_editor.SaveFile(saveas=True, ipage=ipage)

    def set_window_title(self):
        super(ifigure_app, self).set_window_title()
        title = self.GetTitle()
        if title.endswith('*'):
            title = title[:-1]
        name = self.proj.getvar('filename')
        if name is None:
            name = 'untitled'
        else:
            name = os.path.basename(name)
        self.write_launcher_file(name)

        if len(title) > 0:
            xxx = ['piScope', name, title]
        else:
            xxx = ['piScope', name]
        title = ':'.join(xxx)
        if not self.proj.get_saved():
            title = title + '*'

        from ifigure.utils.mp_tarzip import MPTarzip
        if not MPTarzip().isReady():
            title = title + ' (save project in progress)'
        # else:
        #    MPTarzip.lc.release()
        self.SetTitle(title)

    def set_proj_saved(self, value):
        #import traceback
        # traceback.print_stack()
        proj = self.proj
        if not value:
            if proj.get_saved():
                proj.set_saved(value)
                self.set_window_title()
        else:
            if not proj.get_saved():
                proj.set_saved(value)
                self.set_window_title()

    def set_filename_2_window_title(self):
        self.set_window_title()
        return
#        name = self.proj.getvar('filename')
#        if name is None: name = 'untitled'
#        title=self.GetTitle()
#        arr = title.split(':')
#        arr[1] = os.path.basename(name)
#        self.SetTitle(':'.join(arr))

    def onSaveAs(self, e):
        from ifigure.utils.mp_tarzip import local_lc
        if not local_lc.acquire(False):
            ret = dialog.message(
                self, 'Previous save job is still running.\n If this is an error, choose unlock. ', 'Please wait (locked)', 2,
                labels=["OK", "Unlock"])
            if ret == 'cancel':
                local_lc.release()
            return

        opath = self.proj.getvar("filename")
        owdir = self.proj.getvar("wdir")

        try:
            odir = os.getcwd()
        except FileNotFoundError:
            os.chdir(os.path.expanduser("~"))
            odir = os.getcwd()

        if odir.startswith(owdir):
            # if current dir is under wdir. we move to home
            # since saveas will change wdir and wdir will be gone
            os.chdir(os.getenv("HOME"))

        self.save_gui_setting()

        try:
            def_path = os.path.dirname(opath)
        except BaseException:
            def_path = os.getcwd()
        path = dialog.write(parent=self, message="Enter Project File Name",
                            defaultfile=os.path.join(def_path, '.pfz'))

        if path != '':
            if path[-4:] != '.pfz':
                path = path + '.pfz'
            print(("saving to " + path))
            self.proj_tree_viewer.ch.savetofile(filename=os.path.join(
                self.proj.getvar("wdir"), '.command_history'))

            success = self.proj.SaveToFile(path, opath=opath)
        else:
            success = False

        if success:
            #           self.update_save_project_menu(True)
            #           self.set_filename_2_window_title()
            nwdir = self.proj.getvar("wdir")
            self.script_editor.switch_wdir(owdir, nwdir)
            self.set_proj_saved(True)

            self.read_recent_files()
            if not path in RECENT_FILE:
                RECENT_FILE.append(path)
                self.write_recent_files()
        else:
            local_lc.release()

        # if current directory does not exist. move to the home

    def onSaveFile(self, e=None, saveas=False):
        self.script_editor.SaveFile(saveas)

    def onCloseProject(self, e=None):
        # file date check should be added future....
        for v in [v for v in self.viewers]:
            if v is not self:
                bk = v.book
                v.Close()
                while bk.isOpen:
                    time.sleep(0.1)
#                  wx.Yield()
        self.canvas.unselect_all()
        self.proj.CloseProject()
        self.script_editor.close_all_pages()
        self.onTD_CloseFile()  # close editor
        self._set_proj(None)
        self.ipage = -1
        self.proj_tree_viewer.update_widget()
        self._force_layout()

#    def onPreference(self, e):
#        from ifigure.widgets.dlg_preference import dlg_preference

#        dlg_preference(components, self)
#        for c in components:
#            c.save_setting()
        #dprint1('preference panel is not yet implemented')

    def onAbout(self, e):
        import ifigure
        from ifigure.ifigure_config import icondir as path
        icon_path = os.path.join(path, 'app_logo_small.png')
        try:
            info = wx.AboutDialogInfo()
            use_adv = False
        except BaseException:
            info = wx.adv.AboutDialogInfo()
            use_adv = True

        info.SetIcon(wx.Icon(icon_path, wx.BITMAP_TYPE_PNG))
        info.SetName('piScope')
        from ifigure import __version__
        info.SetVersion(__version__)
        info.SetDescription(
            'Python scripting workbench for \n   - browing MDSplus Data.\n   - analysing experiemntal data.\n   - running simulation codes.\n        ...\n')
        info.SetCopyright('(C) 2012 - 2019 S. Shirawa')
        info.SetWebSite('http://piscope.psfc.mit.edu/')

        if use_adv:
            wx.adv.GenericAboutBox(info, self)
        else:
            wx.AboutBox(info, self)

    def onClearCommandHistory(self, e):
        self.proj_tree_viewer.ch.clear_text()

    #
    #   Edit menue
    #
    def onCut(self, e):
        fc = self.FindFocus()

        while fc is not None:
            if (fc is self.canvas.canvas or
                    fc is self.property_editor):
                super(ifigure_app, self).onCut(e)
                self.proj_tree_viewer.update_widget()
                break
            # if isinstance(fc, ProjTreeViewer):
            #    fc.Cut()
            #    break
            if isinstance(fc, SimpleShell):
                fc.Cut()
                break
            elif isinstance(fc, CommandHistory):
                fc.Cut()
                break
            elif isinstance(fc, wx.stc.StyledTextCtrl):
                fc.Cut()
                break
            elif isinstance(fc, wx.TextCtrl):
                fc.Cut()
                break
            else:
                pass

            fc = fc.GetParent()

    def onCopy(self, e):
        fc = self.FindFocus()

        while fc is not None:
            if (fc is self.canvas.canvas or
                    fc is self.property_editor):
                super(ifigure_app, self).onCopy(e)
#               self.proj_tree_viewer.update_widget()
                break
            elif fc is self.proj_tree_viewer.tree:
                if e.GetEventObject() != self:
                    check = self.proj_tree_viewer.copy_tree_item()
                    if not check:
                        dlg = wx.MessageDialog(self,
                                               'Selected Item can not be copyed',
                                               'Error',
                                               wx.OK)
                        ret = dlg.ShowModal()
                        dlg.Destroy()
                    else:
                        self.set_status_text('Copy (tree item)', timeout=3000)
#                    print 'copy in tree itme is selected from menu'
                break
            elif isinstance(fc, SimpleShell):
                fc.Copy()
                break
            elif isinstance(fc, CommandHistory):
                fc.Copy()
                break
            elif isinstance(fc, wx.stc.StyledTextCtrl):
                fc.Copy()
                break
            elif isinstance(fc, wx.TextCtrl):
                fc.Copy()
                break
            else:
                pass

            fc = fc.GetParent()
            #        if fc is self.canvas.canvas:
            #           super(ifigure_app, self).onCopy(e)
            #           self.proj_tree_viewer.update_widget()
            #        if isinstance(fc, wx.stc.StyledTextCtrl):
            #           fc.Copy()
            #        if isinstance(fc, ProjTreeViewer):
            #           fc.Copy()

    def onPaste(self, e):
        fc = self.FindFocus()
#        print fc.GetParent(), fc
        while fc is not None:
            if (fc is self.canvas.canvas or
                    fc is self.property_editor):
                super(ifigure_app, self).onPaste(e)
                self.proj_tree_viewer.update_widget()
                break
            elif fc is self.proj_tree_viewer.tree:
                if e.GetEventObject() != self:
                    check = self.proj_tree_viewer.paste_tree_item()
                    if not check:
                        dlg = wx.MessageDialog(self,
                                               'Selected Item can not be pasted',
                                               'Error',
                                               wx.OK)
                        ret = dlg.ShowModal()
                        dlg.Destroy()
                    else:
                        self.set_status_text('Paste (tree item)', timeout=3000)
                break
            elif isinstance(fc, SimpleShell):
                fc.Paste()
                break
            elif isinstance(fc, wx.stc.StyledTextCtrl):
                fc.Paste()
                break
            elif isinstance(fc, wx.TextCtrl):
                fc.Paste()
                break
            else:
                pass

            fc = fc.GetParent()

    def onAppWindowClose(self, e):

        from ifigure.utils.mp_tarzip import local_lc
        if not local_lc.acquire(False):
            ret = dialog.message(
                self, 'Save job is running. Application will close after \ndata is saved.', 'Please wait', 0)
            local_lc.acquire()
        local_lc.release()

        filename = self.proj.eval("filename")
        if filename is not None:
            print("#### piScope Applicaiton window closed: project: "
                  + os.path.basename(filename))
        else:
            print("#### piScope Applicaiton window closed: project: project not saved")

        from ifigure.widgets.debugger import is_waiting
        if is_waiting():
            self.script_editor.d_panel.StopDebugger()
        sys.stdout = self.stdfiles[0]
        sys.stderr = self.stdfiles[1]

        self.shell.write_history()
        self.write_geom_file()

        for v in [v for v in self.viewers]:
            if v is not self:
                bk = v.book
                v.clear_history()
                v.Close()
                while bk.isOpen:
                    time.sleep(0.1)

        self.viewer = []
        self.clear_history()

        # close script editor if it is detached
        try:
            se = self.script_editor.GetTopLevelParent()
            if se != self:
                se.Close()
        except PyDeadObjectError:
            pass

        # stop file system checking from script edtior
        self.Unbind(ifigure.events.TD_EVT_FILESYSTEMCHANGED)
        if self.proj is not None:
            self.proj.CloseProject()

        self.save_pref()

        try:
            self.tipw.Destroy()
        except PyDeadObjectError:
            pass
        try:
            self.logw.Destroy()
        except PyDeadObjectError:
            pass

        self.proj.destroy()

        import multiprocessing
        children = multiprocessing.active_children()
        if len(children) > 0:
            dprint1("terminating active children : " + str(len(children)))
            for x in children:
                x.terminate()

        # close script editor if it is detached
        try:
            from ifigure.mto.hg_support import diffwindow
            if diffwindow is not None:
                diffwindow.Destroy()
        except PyDeadObjectError:
            pass

        self.Destroy()  # is this necessary to make sure app closes??? (2015.10)
        e.Skip()

#    def onQuit(self, e=None):
#        self.Close()

    def onShellEnter(self, evt):
        self.set_proj_saved(False)
        self.proj_tree_viewer.get_shellvar_viewer().update(self.shell)
        self.proj_tree_viewer.get_var_viewer().update()

    def write_geom_file(self):
        size = self.GetSize()
        sh = self.gui_tree.get_showhide()
        pos = self.gui_tree.get_sashposition()
        print(('editor detached', not self.isEditorAttached()))
        val = {"version": 0, "size": size, "sh": sh, "pos": pos,
               'editor_detached': not self.isEditorAttached(),
               'RECENT_FILE': RECENT_FILE}

        from ifigure.ifigure_config import geom_file
        fid = open(geom_file, 'wb')
        pickle.dump(val, fid)
        fid.flush()
        fid.close()

    def read_geom_file(self):
        from ifigure.ifigure_config import geom_file
        if os.path.exists(geom_file):
            fid = open(geom_file, 'rb')
            val = pickle.load(fid)
            fid.close()
            return val
        else:
            # this is temporary geometry data (used only first launch)
            return {'sh': [True, False, True, True, True],
                    'size': wx.Size(800, 500),
                    'pos': [200, 248, 300]}

    def read_recent_files(self):
        val = self.read_geom_file()
        if 'RECENT_FILE' in val:
            for item in val['RECENT_FILE']:
                if not item in RECENT_FILE and os.path.exists(item):
                    RECENT_FILE.append(item)

    def write_recent_files(self):
        val = self.read_geom_file()
        val['RECENT_FILE'] = RECENT_FILE

        from ifigure.ifigure_config import geom_file
        fid = open(geom_file, 'wb')
        pickle.dump(val, fid)
        fid.flush()
        fid.close()

    def draw(self, *args, **kargs):
        if self.aviewer == self:
            super(ifigure_app, self).draw(*args, **kargs)
        else:
            self.aviewer.draw(*args, **kargs)

    def draw_all(self, *args, **kargs):
        if self.aviewer == self:
            super(ifigure_app, self).draw_all(*args, **kargs)
        else:
            self.aviewer.draw_all(*args, **kargs)

    def show_page(self, ipage=0, last=False, first=False):
        if self.aviewer == self:
            super(ifigure_app, self).show_page(ipage=ipage,
                                               last=last, first=first)
        else:
            self.aviewer.show_page(ipage=ipage,
                                   last=last, first=first)
        self.proj_tree_viewer.update_widget(no_set_selection=True)

    def add_page(self, *args, **kargs):
        if self.aviewer == self:
            return super(ifigure_app, self).add_page(*args, **kargs)
        else:
            return self.aviewer.add_page(*args, **kargs)

    def get_page(self, *args, **kargs):
        if self.aviewer == self:
            return super(ifigure_app, self).get_page(*args, **kargs)
        else:
            return self.aviewer.get_page(*args, **kargs)

    def del_page(self, *args, **kargs):
        if self.aviewer == self:
            return super(ifigure_app, self).del_page(*args, **kargs)
        else:
            return self.aviewer.del_page(*args, **kargs)

    def set_selection(self, *args, **kargs):
        if self.aviewer == self:
            return super(ifigure_app, self).set_selection(*args, **kargs)
        else:
            return self.aviewer.set_selection(*args, **kargs)

    def get_axes(self, *args, **kargs):
        if self.aviewer == self:
            return super(ifigure_app, self).get_axes(*args, **kargs)
        else:
            return self.aviewer.get_axes(*args, **kargs)

    def set_axes(self, *args, **kargs):
        if self.aviewer == self:
            return super(ifigure_app, self).set_axes(*args, **kargs)
        else:
            return self.aviewer.set_axes(*args, **kargs)

    def open_newbook_in_newviewer(self, viewer_class):
        book = self.proj.onAddBook()
        i_page = book.add_page()
        page = book.get_page(i_page)
        page.add_axes()
        page.realize()
#       page.set_section(*args)

        from ifigure.widgets.book_viewer import BookViewer
        v = viewer_class(self, wx.ID_ANY, book.get_full_path(),
                         book=book)
        self.viewers.append(v)
        if self.aviewer is None:
            self.aviewer = v
        self.proj_tree_viewer.update_widget()
        return book, v

    def open_book_in_newviewer(self, viewer_class, book, ipage=0, kwargs={}):
        if book.num_page() == 0:
            print("***No pages***")
            return
        v = viewer_class(self, wx.ID_ANY, book.get_full_path(),
                         book=book, **kwargs)
#        v.adjust_frame_size()
#        elif evt.viewer == 'mdsscope':
#           v = MDSScope(self, wx.ID_ANY, book.get_full_path(),
#                       book=book, connection_type=evt.param)
        self.viewers.append(v)
        self.proj_tree_viewer.update_widget()
        v.show_page(ipage)
        if self.aviewer is None:
            self.aviewer = v
        return v

    def open_book_in_appwindow(self, book, ipage=0):
        if self.book is not None:
            self.close_figurebook()
        self.book = book
        self.ipage = 0
        self.book.set_open(True)
        self.viewers.append(self)
        self.gui_tree.show_toggle_menu(self.panel1)
        self.gui_tree.show_toggle_menu(self.property_editor)
        if self.aviewer is None:
            self.aviewer = self
        self.gui_tree.rebuild_menu()
        self.gui_tree.toggle_panel(self.panel1, True)
        self.proj_tree_viewer.update_widget()
        self.open_book()

        BookViewerFrame.show_page(self, ipage)
        [x.Enable(True) for x in self._figure_mnis]
        [v.adjust_attach_menu() for v in self.viewers]
        if self.aviewer is None:
            self.aviewer = v

        self.deffered_force_layout()
        self.draw_all()

    def make_attached_script_editor(self):
        self.script_editor = ScriptEditor(self.panel2)
        sizer = self.panel2.GetSizer()
        sizer.Add(self.script_editor, 1, wx.EXPAND | wx.ALL)

    def onDetachFigure(self, evt):
        book = self.book
        ipage = self.ipage
        aviewer = self.aviewer
        self.close_figurebook()
        viewer = self.open_book_in_newviewer(BookViewer, book, ipage=ipage)
        if aviewer is self:
            self.aviewer = viewer

    def attach_editor(self, open_editor=True):
        self.script_editor.Reparent(self.panel2)
        self.gui_tree.show_toggle_menu(self.panel2)
        self.gui_tree.rebuild_menu()
        self.gui_tree.toggle_panel(self.panel2, open_editor)
        sizer = self.panel2.GetSizer()
        sizer.Add(self.script_editor, 1, wx.EXPAND | wx.ALL)
        sizer.Layout()

    def onDetachEditor(self, evt):
        # evt can be None
        #        self._editor_mni.Enable(False)
        self.gui_tree.hide_toggle_menu(self.panel2)
        self.gui_tree.rebuild_menu()
        self.gui_tree.toggle_panel(self.panel2, False)

        self.panel2.GetSizer().Detach(self.script_editor)
#        sel, contents = self.script_editor.get_all_data()
#        self.script_editor.Destroy()
        from ifigure.widgets.script_editor import ScriptEditorFrame
        frame = ScriptEditorFrame(self)
        frame.set_script_editor(self.script_editor)

#        self.script_editor.set_all_data(sel, contents)

    def onCloseFigure(self, evt):
        if self.book._keep_data_in_tree:
            self.close_figurebook()
        else:
            self.onCloseFigureDeleteBook(evt)

    def onCloseFigureDeleteBook(self, evt=None):
        b = self.book
        self.close_figurebook()
        b.destroy()
        self.proj_tree_viewer.update_widget()

    def close_figurebook(self):
        ##
        # close figurebook opened in main window...
        ##

        # self.viewers.remove(self)

        # if len(self.viewers) == 0:
        #    self.aviewer = None
        # elif self.aviewer == self:
        #    self.aviewer = self.viewers[0]

        self.book.set_open(False)
        s = self.canvas.GetClientSize()
        self.book._screen_size = (s[0], s[1])

        self._go_closed_figure_mode()

        self.set_window_title()
        [x.Enable(False) for x in self._figure_mnis]
        [v.adjust_attach_menu() for v in self.viewers]

    def _go_closed_figure_mode(self):
        if not self in self.viewers:
            return  # already in closed_figure_mode
        self.viewers.remove(self)
        # print len(self.viewers)
        if len(self.viewers) == 0:
            self.aviewer = None
        elif self.aviewer == self:
            self.aviewer = self.viewers[0]

        self.gui_tree.toggle_panel(self.panel1, False)
        self.gui_tree.hide_toggle_menu(self.panel1)
        self.gui_tree.hide_toggle_menu(self.property_editor)
        self.gui_tree.rebuild_menu()
        self.proj_tree_viewer.update_widget()
        self.canvas.set_figure(None)
        self.book = None

    def run_text(self, txt, no_exec=False):
        if not no_exec:
            self.shell.lvar['command'] = txt + '\n'
            self.shell.execute_text(txt)
        else:
            self.shell.write(txt)

    def _set_proj(self, proj):
        self.proj = proj
        if self.proj is not None:
            self.proj.set_app(self)
        if hasattr(self, 'shell'):
            self.shell.set_proj(self.proj)

#    def update_save_project_menu(self, value):
#        pass
#        for v in self.viewers:
#            if hasattr(v, 'save_project_menu'):
#                v.save_project_menu.Enable(value)
#        self.save_project_menu.Enable(value)

    def update_open_filelist(self, list):
        '''
        save file names opend in script editor
        it saves only the files under wdir.
        '''
        nlist = []
        if self.proj is not None:
            wdir = self.proj.getvar("wdir")
            if wdir is None:
                return
        else:
            return
        for item in list:
            c1 = os.path.commonprefix([wdir, item])
            if c1 == wdir:
                nlist.append((item, 1))
            else:
                nlist.append((item, 0))
        dot_file = os.path.join(wdir, '.opened_file')
        fid = open(dot_file, 'w')
        for item, flag in nlist:
            if flag:
                relpath = os.path.relpath(item, wdir)
            else:
                relpath = item
            if hasattr(item, 'td'):
                relpath = relpath + '\t' + item.td
            fid.write(relpath + '\n')
        fid.close()

    def onTD_EvtUndo(self, evt):
        fc = self.FindFocus()
        if fc is not None:
            if isinstance(fc, wx.stc.StyledTextCtrl):
                fc.Undo()
                return
            elif isinstance(fc, SimpleShell):
                fc.Undo()
                return

        td = GlobalHistory().get_history(evt.GetEventObject()).undo()
        if isinstance(td, FigObj):
            #           print 'change happens under', td.get_full_path()
            book = td.get_root_figobj()
            viewer = self.find_bookviewer(book)
            if viewer is not None:
                self.aviewer = viewer
#              viewer.canvas.draw()
#              viewer.canvas.refresh_hl_idle()
                viewer.property_editor.update_panel()

    def onTD_EvtRedo(self, evt):
        fc = self.FindFocus()
        if fc is not None:
            if isinstance(fc, wx.stc.StyledTextCtrl):
                fc.Redo()
                return
            elif isinstance(fc, SimpleShell):
                fc.Redo()
                return

#       td=self.history.redo()
        td = GlobalHistory().get_history(evt.GetEventObject()).redo()
        if isinstance(td, FigObj):
            #           print 'change happens under', td.get_full_path()
            book = td.get_root_figobj()
            viewer = self.find_bookviewer(book)
            if viewer is not None:
                self.aviewer = viewer
#              viewer.canvas.draw()
#              viewer.canvas.refresh_hl_idle()
                viewer.property_editor.update_panel()

    def onTD_EditFile(self, evt):
        if self.helper.setting['use_editor']:
            root = os.path.dirname(os.path.dirname(ifigure.__path__[0]))
            bin = os.path.join(root, 'bin')
            txt = self.helper.setting['editor']
            txt = txt.replace('{bin}', bin)
            txt = txt.format(evt.file)
            dprint1(txt)
            os.system(txt)
        else:
            readonly = evt.readonly if hasattr(evt, 'readonly') else False
            self.script_editor.OpenFile(file=evt.file, readonly=readonly)
            if self.isEditorAttached():
                self.gui_tree.set_showhide([True], self.panel2)
                self.gui_tree.show_toggle_menu(self.panel2)
                self.gui_tree.rebuild_menu()
                self.gui_tree.update_check()
                self.gui_tree.set_splitters()

            list = self.script_editor.get_filelist()
            self.update_open_filelist(list)

    def onTD_CloseFile(self, evt=None):
        list = self.script_editor.get_filelist()
        if len(list) == 0:
            if self.isEditorAttached():
                self.gui_tree.set_showhide([False], self.panel2)
                self.gui_tree.hide_toggle_menu(self.panel2)
                self.gui_tree.rebuild_menu()
                self.gui_tree.update_check()
                self.gui_tree.set_splitters()
            else:
                self.script_editor.GetTopLevelParent().Close()
        self.update_open_filelist(list)
        self.Layout()

    def onTD_Changed(self, evt):
        from ifigure.events import changed_in_queue

        figobj = evt.GetTreeDict()

        def update_proj_viewer(evt, td=figobj, time=evt.time):
            ifigure.events.changed_in_queue = ifigure.events.changed_in_queue - 1
            if ifigure.events.changed_in_queue > 0:
                dprint2('more changed events are coming, let skip')
                return
            self.Unbind(wx.EVT_IDLE)
            dprint2('processing changed event')
            self.proj_tree_viewer.update_widget_request2(time)
            dprint2('update_widget_done')
            if (isinstance(td, FigObj) and
                td.get_figbook() is not None and
                    td.get_figbook().isOpen):
                viewer = self.find_bookviewer(td.get_figbook())
                if viewer is not None:
                    viewer.property_editor.update_panel_request2(time)
#            ifigure.events.changed_in_queue = False
#            if contents_update:
#            self.proj_tree_viewer.update_content_widget(td)

        self.Bind(wx.EVT_IDLE, update_proj_viewer)

    def onTD_Var0Changed(self, evt):
        self.proj_tree_viewer.update_content_widget(evt.GetTreeDict())

    def onTD_Selection(self, evt):
        td = evt.GetTreeDict()
        w = evt.GetEventObject()
        fc = self.FindFocus()
        if isinstance(td, FigObj):
            book = td.get_root_figobj()
            viewer = self.find_bookviewer(book)
            if viewer is not None:
                self.aviewer = viewer
            if viewer is not None:
                if w != viewer.property_editor:
                    viewer.property_editor.onTD_Selection(evt)
                if w != viewer.canvas:
                    viewer.canvas.onTD_Selection(evt)
                # each bookviewer may want to get this event..
                viewer.onTD_SelectionInFigure(evt)
        if w != self.proj_tree_viewer:
            self.proj_tree_viewer.onTD_Selection(evt)
        try:
            if w != self.hdf_export_window:
                if self.hdf_export_window is not None:
                    self.hdf_export_window.onTD_Selection(evt)
        except PyDeadObjectError:
            self.hdf_export_window = None

#       this is added to preserve focus.
#       not a perfect solution.
        if fc is not None:
            fc.SetFocus()
            try:
                fc.SetFocus()
            except PyDeadObjectError:
                pass

    def onTD_Replace(self, evt):
        super(ifigure_app, self).onTD_Replace(evt)
        history = GlobalHistory().get_history(evt.GetEventObject())
        history.replace_artist(evt.a1, evt.a2)
#        print 'onTD_replace ifigure_app'

    def onTD_ThreadStart(self, evt):
        import time
#        print 'thread start event...', threading.current_thread().name
        if self.logw.check_thread_can_start(evt.thread):
            evt.dt._status = 'running ...'
            # self.proj_tree_viewer.update_widget()
            self.logw.watch_thread(evt)
#            evt.thread.start()
#            print evt.dt._status
#            while evt.thread.is_alive():
#               time.sleep(1)
#               print 'waiting'
        self.proj_tree_viewer.update_widget(no_set_selection=True)

    def onTD_RangeChanged(self, evt):
        # this is for mouse drag in colorbar
        # looks like generating unnecessary draw....

        dprint1('process range event')

        def rebind_range_changed(evt):
            #            print 'rebinding'
            self.Bind(ifigure.events.TD_EVT_RANGECHANGED,
                      self.onTD_RangeChanged)
            self.Unbind(wx.EVT_IDLE)
        self.Unbind(ifigure.events.TD_EVT_RANGECHANGED)
        self.Bind(wx.EVT_IDLE, rebind_range_changed)

        book = evt.GetTreeDict().get_figbook()
        viewer = self.find_bookviewer(book)
#        viewer.canvas.onTD_RangeChanged(evt)
        viewer.property_editor.update_panel()

    def onTD_OpenBook(self, evt):

        book = evt.GetTreeDict()
        viewer_class = evt.viewer
        viewer = self.open_book_in_newviewer(viewer_class, book,
                                             ipage=evt.ipage, kwargs=evt.kwargs)
        if viewer is None:
            print(self.proj.getvar("filename"))
            return

        if evt.isPropShown is not None:
            if viewer.isPropShown() != evt.isPropShown:
                wx.CallAfter(viewer.toggle_property)
        if evt.size is not None:
            wx.CallAfter(viewer.SetSize, evt.size)
        if evt.pos is not None:
            from ifigure.utils.cbook import get_current_display_size
            x0, y0, xd, yd = get_current_display_size(self)
            if (evt.pos[0] < 0.9 * xd and
                    evt.pos[1] < 0.9 * yd):
                wx.CallAfter(viewer.SetPosition, evt.pos)
        return viewer

    def onTD_CloseBook(self, evt):
        if evt.GetEventObject() in self.viewers:
            self.viewers.remove(evt.GetEventObject())
        else:
            dprint1('not found in current viewer list', evt.GetEventObject())
        if (self.aviewer == evt.GetEventObject() and
                len(self.viewers) != 0):
            self.aviewer = self.viewers[0]
        if len(self.viewers) == 0:
            self.aviewer = None
        self.proj_tree_viewer.update_widget()

        if not self.IsShown() and len(self.viewers) == 0:
            wx.CallAfter(self.onQuit)

    def onTD_CloseBookRequest(self, evt):
        evt.GetEventObject().Close()

    def onTD_FileChanged(self, evt):
        self.script_editor.onTD_FileChanged(evt)

    def onTD_PageShown(self, evt):
        td = evt.GetTreeDict()
        self.proj_tree_viewer.set_td_selection(td)

    def onTD_ShowPage(self, evt):
        if not evt.BookViewerFrame_Processed:
            super(ifigure_app, self).onTD_ShowPage(evt)
            evt.BookViewerFrame_Processed = True
            evt.SetEventObject(self)
        return

    def onTD_Evt(self, evt):
        if evt.code == 'Rename':
            for v in self.viewers:
                v.set_window_title()

    def onPV_DrawRequest(self, evt):
        td = evt.GetTreeDict()
        book = td.get_root_figobj()
        viewer = self.find_bookviewer(book)
        if viewer is None:
            return
        self.aviewer = viewer
#        if evt.wait_idle:
#           viewer.canvas.refresh_hl_idle()
#        else:
        if evt.time > viewer.last_draw_time():
            viewer.draw(refresh_hl=evt.refresh_hl)

    def onPV_AddFigobj(self, evt):
        td = evt.GetTreeDict()
        book = td.get_root_figobj()
        viewer = self.find_bookviewer(book)
        if viewer is None:
            return
        self.aviewer = viewer
        page = book.get_page(viewer.ipage)
        if page.is_descendant(td) or page == td:
            viewer.draw()
            viewer.Raise()

    def onPV_DeleteFigobj(self, evt):
        td = evt.GetTreeDict()
        book = td.get_root_figobj()
        viewer = self.find_bookviewer(book)

        if viewer is None:
            td.destroy()
            self.proj_tree_viewer.update_widget_request2(evt.time)
            return
        else:
            self.aviewer = viewer
            page = book.get_page(viewer.ipage)
            if isinstance(td, FigPage):
                ipage = book.i_page(td)
                npage = book.num_page()
                if npage == 1:
                    if viewer is not self:
                        viewer.Close()
                        td.destroy()
                    else:
                        # do nothing
                        return
                else:
                    book.i_page(td)
                    viewer.del_page(ipage)
                self.proj_tree_viewer.update_widget_request2(evt.time)
                return
            elif td == book:
                if viewer is not self:
                    viewer.Close()
                    td.destroy()
                    wx.CallAfter(self.proj_tree_viewer.update_widget_request,
                                 delay=500)
                else:
                    self.close_figurebook()
                    td.destroy()
                    wx.CallAfter(self.proj_tree_viewer.update_widget_request,
                                 delay=500)

            else:
                redraw = page.is_descendant(td)
                td.destroy()
                self.proj_tree_viewer.update_widget_request2(evt.time)
                if redraw:
                    viewer.draw()
                    viewer.Raise()

    def onCV_CanvasSelected(self, evt):
        fig_page = evt.GetTreeDict()
        book = fig_page.get_parent()
        viewer = self.find_bookviewer(book)
        if viewer is not None:
            self.aviewer = viewer

    def onRemoteCommand(self, evt):
        server = ifigure.server.Server()
        try:
            response = server.process(evt.command)
        except BaseException:
            import traceback
            traceback.print_exc()
            response = 'failed'
        self.server_response_queue.put(response)

    def use_server(self):
        from six.moves import queue as Queue

        self.server_response_queue = Queue.Queue()
        return self.server_response_queue

    def find_bookviewer(self, book):
        for viewer in self.viewers:
            if viewer.book == book:
                return viewer
        return None

    def goto_no_mainwindow(self):
        if len(self.viewers) == 0:
            return
        if (len(self.viewers) == 1 and
                self.viewers[0] is self):
            return
        self.Hide()

    def set_launcher_file(self, file):
        self._launcher_file = file
        if self.proj is None:
            return
        name = self.proj.getvar('filename')
        if name is None:
            name = 'untitled'
        else:
            name = os.path.basename(name)
        self.write_launcher_file(name)

    def get_launcher_file(self):
        return self._launcher_file

    def write_launcher_file(self, name):
        file = self.get_launcher_file()
        if file is None:
            return

        filename = os.path.join(file, str(os.getpid()))
        fid = open(filename, 'w')
        fid.write(name)
        fid.close()


class MyApp(wx.App):
    def __init__(self, *args, **kwargs):
        self._palettes = weakref.WeakKeyDictionary()
        self.AppWindow = None
        wx.App.__init__(self, *args, **kwargs)

    def OnInit(self, launcher_file=None):
        self._ifig_app = ifigure_app(None, "iScope+:")
        self.SetTopWindow(self._ifig_app)
        self.AppWindow = self._ifig_app

        self._ifig_app.Show(True)

        if "wxMac" in wx.PlatformInfo:
            self._ifig_app.Bind(wx.EVT_MENU, self.MacQuit,
                                id=self.GetMacExitMenuItemId())
        return True

    def add_palette(self, window):
        if not window.GetParent() in self._palettes:
            self._palettes[window.GetParent()] = [window, ]
        else:
            self._palettes[window.GetParent()].append(window)

    def raise_palette(self, window):
        if not window in self._palettes:
            return
        x = self._palettes[window]
        for w in x:
            try:
                w.Raise()
            except BaseException:
                import traceback
                traceback.print_exc()
                pass
        wx.CallAfter(window.Raise)

    def rm_palette(self, window):
        x = self._palettes[window.GetParent()]
        if window in x:
            x.remove(window)
        if len(x) == 0:
            self._palettes[window.GetParent()]

    def clean_palette(self):
        #        from wx._core import _wxPyDeadObject
        dead_keys = [key for key in self._palettes if not key]
#                     if isinstance(key, _wxPyDeadObject)]
        for key in dead_keys:
            del self._palettes[key]
        dead_keys = []
        for key in self._palettes:
            dead_window = []
            for x in self._palettes[key]:
                #               if isinstance(x, _wxPyDeadObject):
                if not x:
                    dead_window.append(x)
            for x in dead_window:
                self._palettes[key].remove(x)
            if len(self._palettes[key]) == 0:
                dead_keys.append(key)
        for key in dead_keys:
            del self._palettes[key]

    def process_child_focus(self, window):
        self.clean_palette()
        for key in self._palettes:
            if key is window:
                for x in self._palettes[key]:
                    if x.IsShown():
                        continue
                    try:
                        x.Show()
                        # x.Raise()
                    except BaseException:
                        pass
            else:
                for x in self._palettes[key]:
                    try:
                        x.Hide()
                    except BaseException:
                        pass

    def get_ifig_app(self):
        return self._ifig_app

    def get_dpi(self):
        system_dpi = self.GetTopWindow().GetDPI()
        dpi = int((system_dpi[0] + system_dpi[1])/2.0)


        dpi = int(72*(dpi/72)**(0.3))
        return dpi

    def MacReopenApp(self):
        if self.GetTopWindow().IsShown():
            self.GetTopWindow().Raise()
#        print 'here'

    def MacQuit(self, evt):
        self.GetTopWindow().Close()



from __future__ import print_function
#
#    merge_sol script.
#
#    it does move all children in work to a new sol,
#    which is named like case1, case2....
#
#
#    this script will be used when model does not have
#    a particular finishg_script.
#
#
#
#    copy this to your model and set finish_script in setting panel
#                          or
#    edit this file and leave the finish_script blank
#
#    this script will be called using RunA(k, worker, sol)
#    writing the data is critical part and user needs to
#    use a lock. Use PySol.aquire_lock() PySol.release_lock()
#
from ifigure.mto.py_code import PySol
sol = args[2]
worker = args[1]
index = args[0]

# sol.aquire_lock()
# print 'worker', worker
# print 'solution', sol


name = 'case' + str(index)
if sol.has_child(name):
    print(('removing ', name))
    sol.get_child(name=name).destroy()
print(('merging solution index=', index))
ps = PySol()
sol.add_child(name, ps)

for name, child in worker.get_children():
    if name == 'parameters':
        child.duplicate(ps, 'parameters')
    else:
        child.move(ps, keep_zorder=False)

# sol.release_lock()



from __future__ import print_function
from scipy.integrate import simpson
from scipy.interpolate import CubicSpline
from scipy.optimize import curve_fit
from ifigure.utils.interpolate import interp1d, interp2d
__version__ = '1.0.0'
__author__ = 'shiraiwa'
'''
nameing convetin for files
'''
import numpy as np


def gname(shot, t):
    return 'g'+str(shot)+'.' + '{:0>5d}'.format(int(t*1000))


def aname(shot, t):
    return 'a'+str(shot)+'.' + '{:0>5d}'.format(int(t*1000))


def kname(shot, t):
    return 'k'+str(shot)+'.' + '{:0>5d}'.format(int(t*1000))


kind = 'cubic'


def ensure_cyclic(pts):
    if pts[0, 0] != pts[-1, 0] or pts[0, 1] != pts[-1, 1]:
        return np.vstack((pts, pts[0]))
    return pts


def XY2pts(X, Y):
    if not isinstance(X, np.ndarray):
        X = np.array(X)
    if not isinstance(Y, np.ndarray):
        Y = np.array(Y)

    pts = np.hstack((X.flatten()[:, np.newaxis],
                     Y.flatten()[:, np.newaxis]))
    return pts


def path_contain(path, xy, check=False, cyclic=True):
    '''
    check if path contains xy based on winding number
    algorithm.

    equivalent to 
      mod(angle(u)-angle(v) - 180, 360) - 180 ???

    if xy is on a point of path, it returns True

    the first point and the last point should be the same
    (ensure_cyclic forces it)a
    '''
    if cyclic:
        path = ensure_cyclic(path)

    dx = path[:, 0] - xy[0]
    dy = path[:, 1] - xy[1]
    d1 = np.sqrt(dx[:-1]**2 + dy[:-1]**2)
    d2 = np.sqrt(dx[1:]**2 + dy[1:]**2)

    d = (dx[:-1]*dy[1:] - dx[1:]*dy[:-1])/d1/d2
    dotp = (dx[:-1]*dx[1:] + dy[1:]*dy[:-1])
    if any(np.isnan(d)):
        # when d1, d2 = 0
        return True

    d = np.arcsin(d)
    d[np.logical_and(dotp < 0, d > 0)] = np.pi - \
        d[np.logical_and(dotp < 0, d > 0)]
    d[np.logical_and(dotp < 0, d < 0)] = -np.pi - \
        d[np.logical_and(dotp < 0, d < 0)]

    xxxx = sum(d)
#    print(np.abs(xxxx)/3.14159/2)
#    if check: return d
    if check:
        return np.abs(xxxx)/np.pi/2
    return -0.001 < np.abs(xxxx)-np.pi*2 < 0.001


def calc_area(path, xy=[0., 0.]):
    '''
    calc area inside path
    '''

    dx = path[:, 0] - xy[0]
    dy = path[:, 1] - xy[1]
    d1 = np.sqrt(dx[:-1]**2 + dy[:-1]**2)
    d2 = np.sqrt(dx[1:]**2 + dy[1:]**2)

    d = (dx[:-1]*dy[1:] - dx[1:]*dy[:-1])
    return np.abs(np.sum(d)/2.)


def area_intersection(p1, p2, area, internal_only=False,
                      closed=True, once=False, return_idx=False):
    '''
    search all possible crosspoint with area on the line connecting p1 and p2
    if internal_only, crosspoint should be between p1 and p2
    p1, p2 : 2D points (x, y)
    area   : [[x1, x2, x3, x4,,,,], [y1, y2, y3,,,,]]
    '''
    x_a = area[0]
    y_a = area[1]
    if closed:
        if x_a[0] != x_a[-1] or y_a[0] != y_a[-1]:
            y_a.append(y_a[0])
            x_a.append(x_a[0])

    p1 = np.array(p1)
    p2 = np.array(p2)

    p = p2 - p1

    x = []
    y = []
    idx = []
    q1 = XY2pts(x_a[:-1], y_a[:-1])
    q2 = XY2pts(x_a[1:],  y_a[1:])
    q0 = q2 - q1
    for k in range(len(area[0])-1):
        #      q1 = np.array([x_a[k], y_a[k]])
        #      q2 = np.array([x_a[k+1], y_a[k+1]])
        #      q  = q2 - q1
        q = q0[k]
        m = np.matrix((-p, q))
        if np.linalg.cond(m) < 1./sys.float_info.epsilon:
            #          ans =  np.array((p1 - q1)*(np.matrix((-p, q))**-1)).flatten()
            ans = np.array((p1 - q1[k])*(np.matrix((-p, q))**-1)).flatten()
        else:
            continue

        if 0. < ans[-1] <= 1.:
            if not internal_only or (internal_only and
                                     (0. < ans[0] <= 1.)):
                #              x.append((q1 + ans[1]*q)[0])
                #              y.append((q1 + ans[1]*q)[1])
                x.append((q1[k] + ans[1]*q)[0])
                y.append((q1[k] + ans[1]*q)[1])
                idx.append(k+ans[1])
                if once:
                    break

    if return_idx:
        return x, y, idx
    else:
        return x, y


'''
find contour
'''


def find_psi_contour(rgrid, zgrid, psirz, rmaxis, zmaxis, psi, return_all=False):
    '''
    return_all : return all segments
    (default)  : return path containing magnetic axis
                 if 0<npsi2psi(psi)<1: this is sufficient to check if it is inside LCFS
    '''
    # from ifigure.utils.geom import path_contain, XY2pts

    try:
        from matplotlib import _cntr as cntr
    except ImportError:
        try:
            from legacycontour import _cntr as cntr
        except ImportError:
            import ifigure.utils.find_contours as cntr

    # if False:
    if True:
        rgrid3 = np.linspace(rgrid[3], rgrid[-3], len(rgrid)*2)
        zgrid3 = np.linspace(zgrid[3], zgrid[-3], len(zgrid)*2)
        f = interp2d(rgrid, zgrid, psirz)
        X3, Y3 = np.meshgrid(rgrid3, zgrid3)
        # psirz3 = [f(X3.flatten(), Y3.flatten()).reshape(X3.shape)
        psirz3 = np.hstack([f(x, y) for x, y in zip(
            X3.flatten(), Y3.flatten())]).reshape(X3.shape)
        # psirz3 = np.transpose(psirz3)
        c = cntr.Cntr(X3, Y3, psirz3)
        rgrid = rgrid3
        zgrid = zgrid3

    else:
        X, Y = np.meshgrid(rgrid, zgrid)
        c = cntr.Cntr(X, Y, psirz)

    res = c.trace(psi)
    nseg = len(res)//2
    segments, codes = res[:nseg], res[nseg:]
    dr = rgrid[1]-rgrid[0]
    dz = zgrid[1]-zgrid[0]

    if return_all:
        return segments
    res = []
    for k, seg in enumerate(segments):
        if path_contain(seg, [rmaxis, zmaxis]):
            return seg


def flux_average(rgrid, zgrid, psirz, rmaxis, zmaxis, q, psi, weight="r"):
    '''
    compute flux surface averaged value of
    quantity q. 
    q: 2D array
    psi: psi on which average is taken

    weight:
       "r": (default) 3D volumetric average.
       "1": compute average as the poloidal path integral w/o the R factor.
       "1/r": original implementation.
    '''
    path = find_psi_contour(rgrid, zgrid, psirz, rmaxis, zmaxis, psi)

    R, Z = np.meshgrid(rgrid, zgrid)
    dr = rgrid[1]-rgrid[0]
    dz = zgrid[1]-zgrid[0]
    dpsidz, dpsidr = np.gradient(psirz)
    br = interp2d(rgrid, zgrid, -dpsidz/dz/R)
    bz = interp2d(rgrid, zgrid, dpsidr/dr/R)
    bp = np.array([np.sqrt(br(x, y)**2+bz(x, y)**2)
                  for x, y in zip(path[:, 0], path[:, 1])]).flatten()

    q = interp2d(rgrid, zgrid, q)
    qq = np.array([q(x, y) for x, y in zip(path[:, 0], path[:, 1])]).flatten()

    # create half grid data
    dl = np.sqrt((path[1:, 0] - path[:-1, 0])**2 +
                 (path[1:, 1] - path[:-1, 1])**2)
    qq = (qq[1:]+qq[:-1])/2
    bp = (bp[1:]+bp[:-1])/2
    r = (path[1:, 0]+path[:-1, 0])/2

    # integration
    if weight == 'r':
        l = np.sum(qq*dl/bp*r)/np.sum(dl/bp*r)
    elif mode == '1':
        l = np.sum(qq*dl/bp)/np.sum(dl/bp)
    elif mode == '1/r':
        l = np.sum(qq*dl/bp/r)/np.sum(dl/bp/r)
    else:
        assert False, "unknown option"

    return l


'''
file below accept gfile object
'''


def limiter_rho(gfile):
    '''
    the distance of limiter normalized minor radius by 
    minor radius as a function of poloidal angle
    '''
    # from ifigure.utils.geom import area_intersection
    ra = gfile.get_contents('table', 'rmaxis')
    za = gfile.get_contents('table', 'zmaxis')
    rb = gfile.get_contents('table', 'rbbbs')
    zb = gfile.get_contents('table', 'zbbbs')
    rl = gfile.get_contents('table', 'xlim')
    zl = gfile.get_contents('table', 'ylim')

    rg = gfile.get_contents('table', 'rgrid')
    zg = gfile.get_contents('table', 'zgrid')
    lll = np.sqrt((rg[-1]-rg[0])**2 + (zg[-1]-zg[0])**2)

    tharr = np.linspace(-np.pi, np.pi, 180)
    tharr = tharr[:-1]

    p1 = [ra, za]
    res = []
    res2 = []
    for th in tharr:
        p2 = [ra + lll*np.cos(th), za + lll*np.sin(th)]
        x, y = area_intersection(p1, p2,
                                 [rl, zl],
                                 internal_only=True, closed=False)
        x2, y2 = area_intersection(p1, p2,
                                   [rb, zb],
                                   internal_only=True, closed=False)
        res.append(np.min(np.sqrt((np.array(x)-ra)**2 + (np.array(y)-za)**2)))
        res2.append(
            np.min(np.sqrt((np.array(x2)-ra)**2 + (np.array(y2)-za)**2)))
    return tharr, np.array(res)/np.array(res2)


def check_path_inside_LCFS(gfile, path):
    # from ifigure.utils.geom import path_contain, XY2pts, calc_area
    rbbbs = gfile.get_contents("table", "rbbbs")
    zbbbs = gfile.get_contents("table", "zbbbs")
    rmaxis = gfile.get_contents("table", "rmaxis")
    zmaxis = gfile.get_contents("table", "zmaxis")
    if (path_contain(path, [rmaxis, zmaxis]) and
            calc_area(XY2pts(rbbbs, zbbbs)) > calc_area(path)):
        return True

    return False


def npsi2path(gfile, npsi, **kargs):
    rgrid = gfile.get_contents("table", "rgrid")
    zgrid = gfile.get_contents("table", "zgrid")
    psirz = gfile.get_contents("table", "psirz")
    zmaxis = gfile.get_contents("table", "zmaxis")
    rmaxis = gfile.get_contents("table", "rmaxis")

    psi = npsi2psi(gfile, npsi)
    path = find_psi_contour(rgrid, zgrid, psirz, rmaxis, zmaxis, psi, **kargs)

    return path


def npsi2psi(gfile, npsi):
    ssimag = gfile.get_contents("table", "ssimag")
    ssibry = gfile.get_contents("table", "ssibry")
    return ssimag + (ssibry-ssimag)*npsi


def psi2npsi(gfile, psi):
    ssimag = gfile.get_contents("table", "ssimag")
    ssibry = gfile.get_contents("table", "ssibry")
    return (psi - ssimag)/(ssibry-ssimag)


def _get_hr_rmid(gfile, fac=4):
    rgrid = gfile.get_contents("table", "rgrid")
    zgrid = gfile.get_contents("table", "zgrid")
    psirz = gfile.get_contents("table", "psirz")
    zmaxis = gfile.get_contents("table", "zmaxis")
    rmaxis = gfile.get_contents("table", "rmaxis")
    npsimid = gfile.get_contents("table", "npsimid")

    idx = np.where(rgrid < rmaxis)[0]
    npsimid2 = npsimid[idx]
    rgrid2 = rgrid[idx]
    idx = np.where(np.logical_and(0 < npsimid2, npsimid2 < 1.))[0]
    if idx[0] != 0:
        idx = [idx[0]-1]+list(idx)
    npsimid2 = npsimid2[idx]
    rgrid2 = rgrid2[idx]

    idx = np.where(rgrid > rmaxis)[0]
    npsimid1 = npsimid[idx]
    rgrid1 = rgrid[idx]
    idx = np.where(np.logical_and(0 < npsimid1, npsimid1 < 1.))[0]
    if idx[-1] != len(rgrid)-1:
        idx = list(idx)+[idx[-1]+1]
    npsimid1 = npsimid1[idx]
    rgrid1 = rgrid1[idx]

#    new_r = (list(np.linspace(rgrid2[0], maxis, len(rgrid2)*fac)[:-1])+
#            list(np.linspace(rmaxis, rgrid1[-1], len(rgrid1)*fac)))
    kkk = 4
    new_r = (list(rgrid2[:-kkk]) +
             list(np.linspace(rgrid2[-kkk+1], rgrid1[kkk], 30)) +
             list(rgrid1[(kkk+1):]))
    new_r = np.array(new_r)
    xxx = list(rgrid2)+[rmaxis]+list(rgrid1)
    yyy = list(npsimid2)+[0.0]+list(npsimid1)
    fx = interp1d(xxx, yyy, kind=kind)
    return new_r, fx(new_r)


def _get_hr_r1(gfile, fac=4):
    rmaxis = gfile.get_contents("table", "rmaxis")
    rgrid, npsi = _get_hr_rmid(gfile, fac=4)
    idx = np.where(rgrid > rmaxis)[0]
#    print npsi[idx], rgrid[idx]
    return npsi[idx], rgrid[idx]


def _get_hr_r2(gfile, fac=4):
    rmaxis = gfile.get_contents("table", "rmaxis")
    rgrid, npsi = _get_hr_rmid(gfile, fac=4)
    idx = np.where(rgrid <= rmaxis)[0]
    return npsi[idx], rgrid[idx]


def _get_enhanced_r1_array(gfile, enhance_center):
    rgrid = gfile.get_contents("table", "rgrid")
    zgrid = gfile.get_contents("table", "zgrid")
    psirz = gfile.get_contents("table", "psirz")
    zmaxis = gfile.get_contents("table", "zmaxis")
    rmaxis = gfile.get_contents("table", "rmaxis")
    npsimid = gfile.get_contents("table", "npsimid")

    idx = np.where(rgrid > rmaxis)[0]
    npsimid = npsimid[idx]
    rgrid = rgrid[idx]
    print('11111')
    idx = np.where(np.logical_and(0 < npsimid, npsimid < 1.))[0]
    if idx[-1] != len(rgrid)-1:
        idx = list(idx)+[idx[-1]+1]

    npsimid = npsimid[idx]
    rgrid = rgrid[idx]
    npsimid = [0]+list(npsimid)
    rgrid = [rmaxis] + list(rgrid)

    if enhance_center > 0:
        k = enhance_center
        x2 = rgrid[0:k]
        y2 = npsimid[0:k]
        fx2 = np.poly1d(np.polyfit(x2, y2, 2))
        xc = np.linspace(rgrid[0], rgrid[k], 15)
        npsic = fx2(xc)
        rgrid = list(xc) + rgrid[(k+1):]
        npsimid = list(npsic) + npsimid[(k+1):]
    return npsimid, rgrid


def npsi2r1(gfile, npsi, enhance_center=4):
    #    npsimid, rgrid = _get_enhanced_r1_array(gfile, enhance_center)
    npsimid, rgrid = _get_hr_r1(gfile, enhance_center)
    rmaxis = gfile.get_contents("table", "rmaxis")
    if npsi < npsimid[0]:
        return rmaxis
    if npsi > np.max(npsimid):
        return np.max(rgrid)

    fx = interp1d(npsimid, rgrid, kind=kind)
    return fx(npsi)


def r12npsi(gfile, r1, enhance_center=4):
    #    npsimid, rgrid = _get_enhanced_r1_array(gfile, enhance_center)
    npsimid, rgrid = _get_hr_r1(gfile, enhance_center)
    if rgrid[-1] < r1:
        return npsimid[-1]
    if rgrid[0] > r1:
        return npsimid[0]
    fx = interp1d(rgrid, npsimid, kind=kind)
    return fx(r1)


def _get_enhanced_r2_array(gfile, enhance_center):
    rgrid = gfile.get_contents("table", "rgrid")
    zgrid = gfile.get_contents("table", "zgrid")
    psirz = gfile.get_contents("table", "psirz")
    zmaxis = gfile.get_contents("table", "zmaxis")
    rmaxis = gfile.get_contents("table", "rmaxis")
    npsimid = gfile.get_contents("table", "npsimid")

    idx = np.where(rgrid < rmaxis)[0]
    npsimid = npsimid[idx]
    rgrid = rgrid[idx]

    idx = np.where(np.logical_and(0 < npsimid, npsimid < 1.))[0]
    if idx[0] != 0:
        idx = [idx[0]-1]+list(idx)
    npsimid = npsimid[idx]
    rgrid = rgrid[idx]
    npsimid = list(npsimid)+[0]
    rgrid = list(rgrid)+[rmaxis]

    if enhance_center > 0:
        k = enhance_center
        x2 = rgrid[-k-1:]
        y2 = npsimid[-k-1:]
        fx2 = np.poly1d(np.polyfit(x2, y2, 2))
        xc = np.linspace(x2[0], x2[-1], 15)
        npsic = fx2(xc)
        rgrid = rgrid[:(-k-2)] + list(xc)
        npsimid = npsimid[:(-k-2)] + list(npsic)
    return npsimid, rgrid


def npsi2r2(gfile, npsi, enhance_center=4):
    #    npsimid, rgrid = _get_enhanced_r2_array(gfile, enhance_center)
    npsimid, rgrid = _get_hr_r2(gfile, enhance_center)
    rmaxis = gfile.get_contents("table", "rmaxis")
    if npsi < npsimid[-1]:
        return rmaxis
    fx = interp1d(np.array(npsimid), np.array(rgrid), kind=kind)
    return fx(npsi)


def r22npsi(gfile, r1, enhance_center=4):
    #    npsimid, rgrid = _get_enhanced_r2_array(gfile, enhance_center)
    npsimid, rgrid = _get_hr_r2(gfile, enhance_center)
    # if npsi < npsimid[-1]: return 0
    if rgrid[-1] < r2:
        return npsimid[-1]
    if rgrid[0] > r2:
        return npsimid[0]
    fx = interp1d(rgrid, npsimid, kind=kind)
    return fx(r1)


def check_rmid_fit(gfile, enhance_center=3):

    rgrid = gfile.get_contents("table", "rgrid")
    zgrid = gfile.get_contents("table", "zgrid")
    npsimid = gfile.get_contents("table", "npsimid")

    npsi = np.linspace(0, 1, 3000)
    r1 = [npsi2r1(gfile, xx, enhance_center=enhance_center) for xx in npsi]
    r2 = [npsi2r2(gfile, xx, enhance_center=enhance_center) for xx in npsi]

    from ifigure.interactive import figure
    v = figure()
    v.plot(rgrid, npsimid, 'bo')
    v.plot(r1,    npsi, 'r')
    v.plot(r2,    npsi, 'k')

#    psi = npsi2psi(gfile, npsi)
#    path = find_psi_contour(rgrid, zgrid, psirz, rmaxis, zmaxis, psi)
#    return np.min(path[:,0])


def npsi2rho(gfile, npsi):
    r1 = npsi2r1(gfile, npsi)
    rmid = np.max(gfile.get_contents("table", "rbbbs"))
    r0 = gfile.get_contents("table", "rmaxis")
    return (r1-r0)/(rmid-r0)


def rho2npsi(gfile, rho):
    rmid = np.max(gfile.get_contents("table", "rbbbs"))
    r0 = gfile.get_contents("table", "rmaxis")
    z0 = gfile.get_contents("table", "zmaxis")

    r = (rmid - r0)*rho+r0
    return r12npsi(gfile, r)


def rz2psi(gfile, r=None, z=None, func=False):
    rgrid = gfile.get_contents("table", "rgrid")
    zgrid = gfile.get_contents("table", "zgrid")
    psirz = gfile.get_contents("table", "psirz")

    fpsi = interp2d(rgrid, zgrid, psirz)
    if func:
        return fpsi
    return fpsi(r, z)


def rz2npsi(gfile, r=0, z=0, func=False):
    rgrid = gfile.get_contents("table", "rgrid")
    zgrid = gfile.get_contents("table", "zgrid")
    psirz = gfile.get_contents("table", "psirz")
    ssimag = gfile.get_contents("table", "ssimag")
    ssibry = gfile.get_contents("table", "ssibry")
    psirz = (psirz - ssimag)/(ssibry-ssimag)

    fpsi = interp2d(rgrid, zgrid, psirz)
    if func:
        return fpsi
    return fpsi(r, z)


def _compute_psit(gfile):
    qpsi = gfile.get_contents("table", "qpsi")
    ssimag = gfile.get_contents("table", "ssimag")
    ssibry = gfile.get_contents("table", "ssibry")

    try:
        from scipy.integrate import cumulative_trapezoid as cumtrapz
    except ImportError:
        from scipy.integrate import cumtrapz

    dpsi = (ssimag - ssibry)/len(qpsi)
    psit = cumtrapz(qpsi)*dpsi
    return np.array([0] + [x for x in psit])
#    return psit


def _compute_npsit(gfile):
    psit = _compute_psit(gfile)
    npsit = (psit-psit[0])/(psit[-1] - psit[0])
    return npsit


def psi2psit(gfile, x):
    '''
    convert poloidal to toroidal flux
    '''
    ssimag = gfile.get_contents("table", "ssimag")
    ssibry = gfile.get_contents("table", "ssibry")

    psit = _compute_psit(gfile)
    psi = np.linspace(ssimag, ssibry, len(psit))

    fx = interp1d(psi, psit)
    return fx(x)


def psit2psi(gfile, x):
    '''
    convert toroidal flux to poloidal flux
    '''
    ssimag = gfile.get_contents("table", "ssimag")
    ssibry = gfile.get_contents("table", "ssibry")

    psit = _compute_psit(gfile)
    psi = np.linspace(ssimag, ssibry, len(psit))

    fx = interp1d(psit, psi)
    return fx(x)


def npsi2npsit(gfile, x, func=False):
    '''
    convert poloidal to toroidal flux
    '''
    ssimag = gfile.get_contents("table", "ssimag")
    ssibry = gfile.get_contents("table", "ssibry")

    if x >= 1.0:
        return 1.
    if x <= 0.0:
        return 0.
    psit = _compute_npsit(gfile)
    psi = np.linspace(0., 1., len(psit))

    fx = interp1d(psi, psit)
    if func:
        return fx
    return fx(x)


def npsit2npsi(gfile, x, func=False):
    '''
    convert toroidal flux to poloidal flux
    '''
    ssimag = gfile.get_contents("table", "ssimag")
    ssibry = gfile.get_contents("table", "ssibry")

    psit = _compute_npsit(gfile)
    psi = np.linspace(0., 1., len(psit))

    fx = interp1d(psit, psi)
    if func:
        return fx
    return fx(x)


def find_flux_surface(gfile, npsi, force_ccw=False, force_cw=False):
    rgrid = gfile.get_contents("table", "rgrid")
    zgrid = gfile.get_contents("table", "zgrid")
    psirz = gfile.get_contents("table", "psirz")
    zmaxis = gfile.get_contents("table", "zmaxis")
    rmaxis = gfile.get_contents("table", "rmaxis")

    psi = npsi2psi(gfile, npsi)
    path = find_psi_contour(rgrid, zgrid, psirz, rmaxis, zmaxis, psi)

    if force_ccw:
        r = path[:, 0] - rmaxis
        z = path[:, 1] - zmaxis
        th = np.arctan2(z, r)
        dth = th[1:] - th[:-1]
        if np.sum(dth < 0) > np.sum(dth > 0):
            path = np.flip(path, 0)
    if force_cw:
        r = path[:, 0] - rmaxis
        z = path[:, 1] - zmaxis
        th = np.arctan2(z, r)
        dth = th[1:] - th[:-1]
        if np.sum(dth < 0) > np.sum(dth > 0):
            path = np.flip(path, 0)

    return path, psi


def npsi2any(gfile, x, unit=None):
    if unit == 'rho':
        return npsi2rho(gfile, x)
    elif unit == 'r1':
        return float(npsi2r1(gfile, x))
    elif unit == 'npsit':
        return float(npsi2npsit(gfile, x))
    elif unit == 'npsit-sqrt':
        return np.sqrt(npsi2npsit(gfile, x))
    elif unit == 'npsi-sqrt':
        return np.sqrt(x)
    else:
        return x


'''
flux coordinate
'''


def compute_theta(gfile, npsi=0, path=None, method='straight_field_line'):
    '''
    compute counter path of npsi and theta on the
    path.
       method = 'straight_field_line'
                'equal_arc_length'
    path starts from outer mid-plane and rotate magnetic
    axis anti-clockwise and has a dimension of [:,2]

    ! both path and theta is NOT cyclic (2016 01 20)
    '''
    def find_cut(path, rmaxis, zmaxis):
        xx = path[:, 0]-rmaxis
        yy = path[:, 1]-zmaxis
        angle = np.arctan2(yy, xx)
        for i in range(len(angle)):
            a2 = np.roll(angle, i)
            if a2[0] >= 0 and a2[-1] < 0:
                return i

        print(angle)
        zz = len(path[:, 0])
        if path[-2, 1] < zmaxis and path[0, 1] >= zmaxis:
            return 0
        for k in range(zz-1):
            km = k-1
            if path[km, 1] < zmaxis and path[k, 1] >= zmaxis:
                return k+1

    rgrid = gfile.get_contents("table", "rgrid")
    zgrid = gfile.get_contents("table", "zgrid")
    psirz = gfile.get_contents("table", "psirz")
    fpol = gfile.get_contents("table", "fpol")
    zmaxis = gfile.get_contents("table", "zmaxis")
    rmaxis = gfile.get_contents("table", "rmaxis")

    if path is None:
        path, psi0 = find_flux_surface(gfile, npsi, force_ccw=True)
        # rotate array so that cut happesn at low field side
        if (path[0, 0] == path[-1, 0] and
                path[0, 1] == path[-1, 1]):
            path = path[:-1, :]
        icut = find_cut(path, rmaxis, zmaxis)
        path = np.roll(path, icut, 0)

        '''
        if path[icut, 0] < rmaxis:
            path = np.flipud(path)
            icut = find_cut(path, rmaxis, zmaxis)

        path = path[:-1,:]
        path = np.roll(path, -icut, 0)
        '''
        # insert outer midplane

        fpsi = interp2d(rgrid, zgrid, psirz)
        psimid = np.array([fpsi(r, zmaxis) for r in rgrid if r > rmaxis])
        rmid = np.array([r for r in rgrid if r > rmaxis])
        f = interp1d(psimid.flatten(), rmid.flatten())
        path = np.vstack(([f(psi0), zmaxis], path, [f(psi0), zmaxis]))

        # path circulate anti-clockwise around magnetic axis
        if path[1, 1] < path[0, 1]:   # z[1] < z[0]
            path = np.array([y for y in reversed(path)])

    # comput magnetic field on path
    f = interp1d(np.linspace(0., 1., len(fpol)), fpol)
    fpol = f(npsi)

    R, Z = np.meshgrid(rgrid, zgrid)
    dr = rgrid[1]-rgrid[0]
    dz = zgrid[1]-zgrid[0]
    dpsidz, dpsidr = np.gradient(psirz)
    br = interp2d(rgrid, zgrid, -dpsidz/dz/R)
    bz = interp2d(rgrid, zgrid, dpsidr/dr/R)

    br = np.array([br(x, y) for x, y in zip(path[:, 0], path[:, 1])]).flatten()
    bz = np.array([bz(x, y) for x, y in zip(path[:, 0], path[:, 1])]).flatten()
#    bp = np.array([np.sqrt(br(x, y)**2+bz(x, y)**2)  for x, y in zip(path[:,0], path[:,1])]).flatten()
    bp = np.sqrt(br*br + bz*bz)
    bt = np.array([fpol/x for x, y in zip(path[:, 0], path[:, 1])])

    # integrate theta on path
    dlsq = (path[1:, 0] - path[:-1, 0])**2 + (path[1:, 1] - path[:-1, 1])**2
    dlsq2 = (path[-1, 0] - path[0, 0])**2 + (path[-1, 1] - path[0, 1])**2
    # dlsq = np.hstack((dlsq, [dlsq2]))

    if method == 'straight_field_line':
        x_half = (path[1:, 0] + path[:-1, 0])/2.
        bp_half = (bp[1:] + bp[:-1])/2.
        bt_half = (bt[1:] + bt[:-1])/2.
        print(dlsq.shape, bt_half.shape, bp_half.shape, x_half.shape)
        theta = np.hstack((0, np.cumsum(np.sqrt(dlsq)*bt_half/bp_half/x_half)))
        theta = theta/theta[-1]*2.*np.pi
        # theta  = theta[:-1]
    elif method == 'equal_arc_length':
        theta = np.hstack((0, np.cumsum(np.sqrt(dlsq))))
        theta = theta/theta[-1]*2.*np.pi
        # theta  = theta[:-1]
    else:
        raise ValueError('unknown method option')

    return path, theta, bt, bz, br


def br_bz_bt2zeta_eta_xi(path, br, bz, bt):
    '''
    br, bz, bt is 1D array
    unit vector for zeta, eta, xi direction

    note : it does not return a covariant/contravariant basis
    '''
    brtz = np.transpose(np.vstack((br, bt, bz)))
    zeta = brtz/np.transpose(np.vstack([np.sqrt(np.sum(brtz**2, 1))]*3))

    bp = np.transpose(np.vstack((br, [0]*len(bz), bz)))
    bp = bp/np.transpose(np.vstack([np.sqrt(np.sum(bp**2, 1))]*3))
#    bp = bp/np.sqrt(np.sum(bp**2))  # unit vector in the poloidal field direction

    # unit vector in the poloidal field direction
    bp2D = bp[:, [0, 2]]
    # approximately this should work?
    if np.sum((path[1:] - path[:-1])*bp2D[1:]) < 0:
        e_theta_hat = -bp  # unit vector in theta
    else:
        e_theta_hat = bp
    # perpendicular to flux surface going outward
    xi = np.transpose(
        np.vstack([e_theta_hat[:, 1], -e_theta_hat[:, 0], [0]*len(bz)]))

    # unit vector in the toroidal field direction
    e_phi_hat = np.transpose(
        np.vstack([[0]*len(bz), [1.]*len(bz), [0.]*len(bz)]))

    # eta is third direction of (zeta, eta, xi) makes right hand
    eta = np.cross(zeta, xi)

    # pitch angle is angle of zeta on e_phi, e_theta plane
    cos = np.sum(zeta*e_phi_hat, 1)     # cos(pitch angle)
    sin = np.sum(zeta*e_theta_hat, 1)   # sins(pitch angle)

#    print np.sqrt(cos**2 + sin**2)
    mu = np.arctan2(sin, cos)
    return zeta, eta, xi, mu


class fit_func(object):
    def __init__(self, coeff):
        self.coeffs = np.array(coeff, copy=False)

    def __call__(self, new_angle):
        new_angle = np.atleast_1d(new_angle)
        for i, ab in enumerate(self.coeffs):
            if i == 0:
                new_r = np.sin(i*new_angle)*ab[0]
                new_r = new_r + np.cos(i*new_angle)*ab[1]
            else:
                new_r = new_r + np.sin(i*new_angle)*ab[0]
                new_r = new_r + np.cos(i*new_angle)*ab[1]
        return new_r


def fit_boundary(gfile=None, path=None, plot=False,
                 npsi=0.995, method='straigh_field_line',
                 fit='spline', maxorder=-1, verbose=True):
    ''''
    fit boundary shape by fourier serious

    f : function to compute R for geometrical theta
    g : function to compute the geometrical theta from flux coordinate
        theta
    >>> f, g, theta0, r, R0, Z0 = fit_boundary(gfile, 
                                               plot=True, 
                                               npsi=0.7, 
                                               method='equal_arc_length')
    >>> tt = g(theta);R = f(tt)
    >>> figure();plot(R*np.cos(tt) + R0, R*np.sin(tt) + Z0)
    '''
    def ft(y, x, n):
        d = x[-1] - x[0]
        dx = np.gradient(x)

        return (simpson(y*np.sin(n*x), x=x)/d,
                simpson(y*np.cos(n*x), x=x)/d)

    def cost_func(x, *params):
        '''
        params = sin_1, sin_2, sin_3, ...,  cos_0, cos_1, ....
        '''
        params = [0] + list(params)
        coeffs = np.array(params).reshape(-1, 2)
        for i, ab in enumerate(coeffs):
            if i == 0:
                new_r = np.sin(i*x)*ab[0]
                new_r = new_r + np.cos(i*x)*ab[1]
            else:
                new_r = new_r + np.sin(i*x)*ab[0]
                new_r = new_r + np.cos(i*x)*ab[1]
        return new_r

    if path is None:
        path, theta, bt, bz, br = compute_theta(gfile, npsi, method=method)

    if (np.sqrt((path[0, 0] - path[-1, 0])**2 +
                (path[0, 1] - path[-1, 1])**2)) < 1e-6:
        # un-cyclic if it is
        # print("DOING THIS")
        path = path[:-1, :]
        theta = theta[:-1]
    if (np.sqrt((path[0, 0] - path[-1, 0])**2 +
                (path[0, 1] - path[-1, 1])**2)) < 1e-6:
        # print("DOING THIS")
        # un-cyclic if it is
        path = path[:-1, :]
        theta = theta[:-1]

    # R0 = (np.max(path[:,0])+np.min(path[:,0]))/2
    # Z0 = (np.max(path[:,1])+np.min(path[:,1]))/2
    R0 = gfile.get_contents("table", "rmaxis")
    Z0 = gfile.get_contents("table", "zmaxis")
    theta0 = np.unwrap(np.arctan2(path[:, 1]-Z0, path[:, 0]-R0))

    r = np.sqrt((path[:, 0]-R0)**2 + (path[:, 1]-Z0)**2)
    if theta0[1] < theta0[0]:
        theta0 = np.flipud(theta0)
        r = np.flipud(r)

    idx = np.where(theta0 < -np.pi)[0]
    if len(idx) > 0:
        k = idx[-1]+1
        theta0 = np.hstack((theta0[k:], theta0[:k]+2*np.pi)).flatten()
        r = np.hstack((r[k:], r[:k])).flatten()

    # make it cyclic again
    theta0 = np.hstack((theta0, [theta0[0]+2*np.pi])).flatten()
    theta = np.hstack((theta, [theta[0]+2*np.pi])).flatten()
    r = np.hstack((r, r[0])).flatten()

    if plot:
        from ifigure.interactive import figure
        v = figure(100)
        v.nsec(2)
        v.plot(theta0, r, 'ro')

    if fit == 'spline':
        f = CubicSpline(theta0, r, bc_type='periodic')
        if plot:
            v.plot(theta0, f(theta0))
    elif fit == 'fourier':
        errors = []
        coeffs = []
        maxorder = min([len(theta0)//2, 15]) if maxorder == -1 else maxorder
        orders = list(range(1, maxorder))

        delta_min = np.mean(np.abs(np.gradient(r)))
        if verbose:
            print('delta_min', delta_min)

        theta00 = np.linspace(0, 2*np.pi, 100)
        for order in orders:
            p0 = np.zeros(2*order-1)
            p0[0] = 1
            ret = curve_fit(cost_func, theta0, r, p0=p0)
            coeff = np.array([0] + list(ret[0])).reshape(-1, 2)
            errors.append(np.sum((cost_func(theta0, *ret[0]) - r)**2))
            coeffs.append(coeff)
            max_error = np.max(np.sqrt((cost_func(theta0, *ret[0]) - r)**2))

            fitted = cost_func(theta00, *ret[0])
            dfitted = np.diff(fitted)
            n_local_minmax = np.sum(dfitted[:-1]*dfitted[1:] < 0)
            ddfitted = np.diff(dfitted)
            n_inflection_points = np.sum(ddfitted[:-1]*ddfitted[1:] < 0)
            if verbose:
                print(max_error, n_local_minmax, n_inflection_points)
            # if n_inflection_points > 4:
            #    errors = errors[:-1]
            #    coeffs = coeffs[:-1]
            #    break
            # if n_local_minmax > 4:
            #    errors = errors[:-1]
            #    coeffs = coeffs[:-1]
            #    break

            if max_error < delta_min:
                break
        '''
        coeff = [ft(r, theta0, i) for i in range(40)]
        errors = [np.sum((fit_func(coeff[:i+1])(theta0) - r)**2) for i in range(30)]
        '''
        idx = np.argmin(errors)
        if verbose:
            print('minimum error found when length = ', idx)
        # f = fit_func(coeff[:idx+1])
        f = fit_func(coeffs[idx])
        delta = np.abs(f(theta0) - r)
        if verbose:
            print('maximum error and average error = ',
                  np.max(delta), np.mean(delta))
        if plot:
            theta00 = np.linspace(0, 2*np.pi, 100)
            v.plot(theta00, f(theta00))
        if plot:
            v.isec(1)
            v.plot(errors)
    else:
        assert False, "unknown fitting metod:" + fit

    g = interp1d(theta, theta0)

    return f, g, theta0, r, R0, Z0


def plot_theta(gfile, method='straight_field_line', rho=[0.1, 0.99], **kargs):
    npsis = np.linspace(rho[0], rho[1], 15)**2
    thetas = np.linspace(0, 2*np.pi, 60)

    resx = []
    resy = []
    for npsi in npsis:
        path, theta, bt, bz, br = compute_theta(gfile, npsi, method=method)
        fx = interp1d(theta, path[:, 0])
        fy = interp1d(theta, path[:, 1])
        x = np.array([fx(th) for th in thetas])
        y = np.array([fy(th) for th in thetas])
        resx.append(x)
        resy.append(y)

    x = np.vstack(resx)
    y = np.vstack(resy)

    from ifigure.interactive import figure
    v = figure()
    v.plot(x, y, **kargs)
    v.plot(np.transpose(x), np.transpose(y), **kargs)

    return x, y, v


def fit_flux_coords(gfile,
                    method='straight_field_line',
                    rho=[0.1, 0.99],
                    plot=True,
                    verbose=True,
                    rsegs=15,
                    psegs=128,
                    **kargs):
    '''
    plot result of fit_boundary
    '''
    npsis = np.linspace(rho[0], rho[1], rsegs)**2
    thetas = np.linspace(0, 2*np.pi, psegs+1)[:-1]

    resx = []
    resy = []

    for npsi in npsis:
        if npsi > 0.25:
            f, g, tetha0, r, R0, Z0 = fit_boundary(gfile, npsi=npsi,
                                                   method=method,
                                                   fit='spline',
                                                   verbose=verbose)
        else:
            f, g, tetha0, r, R0, Z0 = fit_boundary(gfile, npsi=npsi,
                                                   method=method,
                                                   fit='fourier',
                                                   maxorder=4,
                                                   verbose=verbose)

        th = g(thetas)
        r = f(th)
        x = r*np.cos(th) + R0
        y = r*np.sin(th) + Z0

        resx.append(x)
        resy.append(y)

    x = np.vstack(resx)
    y = np.vstack(resy)

    if plot:
        from ifigure.interactive import figure
        v = figure()
        v.plot(x, y, **kargs)
        v.plot(np.transpose(x), np.transpose(y), **kargs)

    npsis = np.hstack([0, npsis])
    return x, y, R0, Z0, npsis


def non_conforming_flux_coords(gfile,
                               method='straight_field_line',
                               rho=[0.1, 0.99],
                               plot=True,
                               verbose=True,
                               rsegs=16,
                               psegs=128,
                               refine=4,
                               **kargs):

    x, y, R0, Z0, npsis = fit_flux_coords(gfile,
                                          method=method,
                                          rho=rho,
                                          rsegs=rsegs*refine+1,
                                          psegs=psegs*refine,
                                          verbose=False,
                                          plot=False)

    average_dist = np.mean(np.sqrt(np.diff(x, 1)**2 + np.diff(y, 1)**2), 1)

    max_exp = int(np.log(average_dist[-1]/average_dist[0])/np.log(2))

    if plot:
        from ifigure.interactive import figure
        v = figure()
        v.plot(x[0::refine, :], y[0::refine], "k-")
        v.plot(np.transpose(x[:, 0::refine]),
               np.transpose(y[:, 0::refine]), "k-")

    seg = 8
    xy_pairs = [(np.array([R0,]),
                 np.array([Z0,]),)]
    for i in np.arange(rsegs*refine+1)[0::refine]:
        exponent = int(
            np.round(np.log(average_dist[i]/average_dist[0])/np.log(2)))

        n_pol = seg*2**exponent  # number of element in poloidal dir.
        if n_pol > psegs:
            n_pol = psegs
        if n_pol > x.shape[1]:
            step = 1
        else:
            step = x.shape[1]//n_pol
        xx = x[i, 0::step]
        yy = y[i, 0::step]

        xy_pairs.append((xx, yy))

        v.plot(np.transpose(x), np.transpose(y), **kargs)
    for xx, yy in xy_pairs:
        v.plot(xx, yy, "o")

    return x, y, xy_pairs, npsis



#  Name   :fig_grp
#
#          group class for fig_obj
#
#
#
#  Author :
#         Syun'ichi Shiraiwa
#  E-mail :
#         shiraiwa@psfc.mit.edu
#
#
#  History :
#
# *******************************************
#     Copyright(c) 2012- S. Shiraiwa
# *******************************************
from ifigure.widgets.canvas.file_structure import *

import logging
import os
import ifigure.utils.cbook as cbook

from ifigure.mto.treedict import TreeDict
from ifigure.mto.fig_obj import FigObj
from ifigure.mto.py_script import PyScript
from ifigure.mto.py_code import PyCode
import ifigure.utils.geom as geom
import ifigure.events

from matplotlib.lines import Line2D
from matplotlib.patches import Rectangle
from matplotlib.transforms import Bbox
import numpy as np

from ifigure.widgets.undo_redo_history import GlobalHistory
from ifigure.widgets.undo_redo_history import UndoRedoArtistProperty
from ifigure.widgets.undo_redo_history import UndoRedoFigobjProperty
from ifigure.widgets.undo_redo_history import UndoRedoFigobjMethod
from ifigure.widgets.undo_redo_history import UndoRedoGroupUngroupFigobj


class FigGrp(FigObj):
    def __new__(cls, *args, **kargs):
        obj = FigObj.__new__(cls, *args, **kargs)
        obj._artist_extent = [None]*4
        return obj

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx = cbook.LoadImageFile(path, 'group.png')
        return [idx]

    @classmethod
    def isFigGrp(self):
        return True

    def isPageObj(self):
        return self._isPageObj

    @classmethod
    def get_namebase(self):
        return 'group'

    @classmethod
    def allow_outside(self):
        ### if fig_obj can be dragged outside of axes ###
        return True

    @classmethod
    def property_in_file(self):
        # define artist property read by mpl.getp and saved in file
        return []

    @classmethod
    def property_in_palette(self):
        # define artist property or _attr shown in palette
        # this function retuns list of key in listparam defined in
        # artist_widgets.py. the 4th field listparam is the name
        # of property read by either mpl.getp (from artist) or
        # figobj.getp (from figobj)
        return []

    @classmethod  # define _attr values to be saved
    def attr_in_file(self):
        return []

    def isDraggable(self):
        val = True
        for name, child in self.get_children():
            val = val & child.isDraggable()
        return val

    def generate_artist(self, *args, **kywds):
        #        print 'entering fig_grp:: generate_artist'
        for name, child in self.get_children():
            if child._suppress:
                child.del_artist(delall=True)
            else:
                child.generate_artist(*args, **kywds)

        ax = self.get_figaxes()
        c = self.get_container()
        is_figgrp = self.get_figaxes() is None

        if self.isempty():
            a = Rectangle((0, 0), 1, 1, facecolor='none', fill=False,
                          edgecolor='none', alpha=0)
            a.figobj = self
            a.figobj_hl = []
            self._artists = [a]
            if is_figgrp:
                c.patches.append(a)
            else:
                c.add_patch(a)
            if ax is not None and ax.get_3d():
                import mpl_toolkits.mplot3d.art3d as art3d
                art3d.patch_2d_to_3d(a)

            figure = self.get_figpage()._artists[0]
            a.set_figure(figure)
            if self.get_figaxes() is not None:
                if len(self.get_figaxes()._artists) != 0:
                    a.axes = self.get_figaxes()._artists[0]

    def del_artist(self, artist=None, delall=False):
        artistlist = self._artists

        for name, child in self.get_children():
            child.del_artist(delall=True)
        if len(artistlist) != 0:
            c = self.get_container()
            is_figgrp = self.get_figaxes() is None

            self.highlight_artist(False, artistlist)
            for a in artistlist:
                if a in c.patches:
                    if is_figgrp:
                        c.patches.remove(a)
                    else:
                        a.remove()

        return super(FigGrp, self).del_artist(artist=artistlist,
                                              delall=delall)

    def highlight_artist(self, val, artist=None):
        figure = self.get_figpage()._artists[0]
        ax = self.get_figaxes()
        if val:
            if len(self._artists[0].figobj_hl) != 0:
                return
            box = self.get_artist_extent_all()
            self._artist_extent = box
            if box[0] is None:
                return
            x = [box[0], box[0], box[1], box[1], box[0]]
            y = [box[3], box[2], box[2], box[3], box[3]]

            hl = Line2D(x, y, marker='s',
                        color='k', linestyle='None',
                        markerfacecolor='k',
                        markeredgewidth=0.5,
                        figure=figure, alpha=0.3)
            figure.lines.append(hl)
            xy = (box[0], box[2])
            w = box[1] - box[0]
            h = box[3] - box[2]
            hlp = Rectangle(xy, w, h, alpha=0.3, facecolor='k',
                            figure=figure)
            if ax is not None:
                x0, y0 = ax._artists[0].transAxes.transform((0, 0))
                x1, y1 = ax._artists[0].transAxes.transform((1, 1))
                bbox = Bbox([[x0, y0], [x1, y1]])
                hlp.set_clip_box(bbox)
                hlp.set_clip_on(True)
            figure.patches.append(hlp)
#           if ax is not None and ax.get_3d():
#               import mpl_toolkits.mplot3d.art3d as art3d
#               art3d.patch_2d_to_3d(hl)
#               if len(ax._artists) != 0:
#                   a.axes = self.get_figaxes()._artists[0]
            self._artists[0].figobj_hl.extend([hl, hlp])
        else:
            if len(self._artists[0].figobj_hl) == 2:
                figure.lines.remove(self._artists[0].figobj_hl[0])
                figure.patches.remove(self._artists[0].figobj_hl[1])
            self._artists[0].figobj_hl = []

    def has_highlight(self, artist):
        return len(artist.figobj_hl) != 0

    def canvas_menu(self):
        m = [("Ungroup",  self.onUngroup, None),
             ("Rasterized All",  self.onSetRasterize, None),
             ("UnRasterized All",  self.onUnsetRasterize, None)]
        return m

    def get_rasterized_action(self, value):
        a = []
        for name, child in self.get_children():
            a.extend(child.get_rasterized_action(value))
        return a

    def onUngroup(self, evt):
        canvas = evt.GetEventObject()
        canvas.ungroup()
#
#  tree viewer menu
#

    def tree_viewer_menu(self):
        return super(FigGrp, self).tree_viewer_menu()

    def canvas_unselected(self):
        self._artist_extent = [None]*4
#
#  Hit test (annotation mode)
#
#   1) nothing is selected
#        no hit happenes on group.
#        if child hit...custom_picker shoud return
#        topgroup if it is not in the selection.
#        if it is in the selection, it should return
#        the second one.. and so on
#        when the hit happens, canvas will highlight the
#        group and therefore
#        it set self._artist_extent
#   2) when self._artist_extent is set
#        it perform FigGrp own picker
#        if it is not area FigGrp keeps is selection
#        otherwise check children..

    def picker_a(self, a, evt):
        if self._artist_extent[0] is None:
            return False, {}

        hit = False
        extra = {}
        hit, extra, type, loc = self.picker_a0(a, evt)
        if not hit:
            return False, {}
        else:
            self._picker_a_type = type
            self._picker_a_loc = loc
            extra = {"child_artist": self._artists[0]}
            # if edge or point was picked, it does not check
            # the children
            if type !=  'area':
                return hit, extra

        children = reversed(
            sorted(
                [(child.getp('zorder'), child)
                 for name, child in self.get_children()]
                , key=lambda x: x[0]
            ))

        for z, child in children:
            if isinstance(child, FigGrp):
                box = child.get_artist_extent_all()
                child._artist_extent = box
                hit2, extra2 = child.picker_a(child._artists[0],
                                              evt)
                child.canvas_unselected()
                if hit2:
                    self._picker_a_type = 'area'
                    self._picker_a_loc = 3
                    self._artist_extent = [None]*4
                    return hit2, {"child_artist": child._artists[0]}
            else:
                for a1 in child.walk_artists():
                    hit2, extra2, type, loc = child.picker_a0(a1, evt)
                    if hit2:
                        self._picker_a_type = 'area'
                        self._picker_a_loc = 3
                        self._artist_extent = [None]*4
                        return hit2, {"child_artist": a1}

        # if it comes here there is two possibility
        # 1) FigGrp was not selected and non of child did not hit
        # 2) FigGrp was hit and non of child did not hit
        return hit, extra

    def get_artist_extent_all(self):
        '''
        retrun the extent of artist in device
        coordinate
        '''
        def _merge(s, x0d, x1d, y0d, y1d):
            s = [min((x0d, s[0])),
                 max((x1d, s[1])),
                 min((y0d, s[2])),
                 max((y1d, s[3])), ]
            return s

        s = [10000, -1, 10000, -1]
        flag = False
        for name, child in self.get_children():
            if isinstance(child, FigGrp):
                x0d, x1d, y0d, y1d = child.get_artist_extent_all()
                if x0d is None:
                    continue
                s = _merge(s, x0d, x1d, y0d, y1d)
                flag = True
            else:
                for a in child._artists:
                    x0d, x1d, y0d, y1d = child.get_artist_extent(a)
                    if x0d is None:
                        continue
                    s = _merge(s, x0d, x1d, y0d, y1d)
                    flag = True
        if not flag:
            return [None]*4
        return s

    def get_artist_extent2(self, a):
        return self.get_artist_extent_all()

    def get_artist_extent(self, a):
        return self._artist_extent

    def picker_a0(self, a, evt):
        ret = super(FigGrp, self).picker_a0(a, evt)
#        print(ret)
        return ret

# dragstart : called when drag starts
    def dragstart_a(self, a, evt):
        self._st_extent = self.get_artist_extent_all()
        redraw = super(FigGrp, self).dragstart_a(a, evt, mode=2)
        for name, child in self.get_children():
            for a1 in child.walk_artists():
                redraw = child.dragstart_a(a1, evt)
            child._drag_hl.set_color('none')
            child._drag_hl.set_markeredgecolor('none')
            child._drag_hl.set_markerfacecolor('none')
        return redraw

# drag : to show transient usr feedback during drag
    def drag_a(self, a, evt, shift=None, scale=None):
        redraw, scale = super(FigGrp, self).drag_a(a, evt,
                                                   shift=shift, scale=scale)

        for name, child in self.get_children():
            for a1 in child.walk_artists():
                redraw, scale = child.drag_a(a1, evt,
                                             shift=shift, scale=scale)
        return redraw, scale

    def drag_a_rm_hl(self, a):
        for name, child in self.get_children():
            for a1 in child.walk_artists():
                child.drag_a_rm_hl(a1)
        super(FigGrp, self).drag_a_rm_hl(a)
# dragdone : finish-up dragging

    def dragdone_a(self, a, evt, shift=None, scale=None):
        shift = evt.guiEvent_memory.ShiftDown()
        redraw, scale0 = super(FigGrp, self).dragdone_a(a, evt,
                                                        shift=shift, scale=scale)

        window = evt.guiEvent_memory.GetEventObject().GetTopLevelParent()
        hist = GlobalHistory().get_history(window)
#        hist = self.get_root_parent().app.history
        h = self.scale_artist(scale0, action=True)
        if len(h) != 0:
            hist.start_record()
            for item in h:
                hist.add_history(item)
            hist.stop_record()
        for name, child in self.get_children():
            for a1 in child.walk_artists():
                child.dragdone_a_clean(a1)
                child.highlight_artist(False, artist=[a1])
        self.highlight_artist(False, artist=[a])
#        self.highlight_artist(True, artist=[a])
        return 0, scale0

    def scale_artist(self, scale, action=True):
        h = []
        for name, child in self.get_children():
            h = h + child.scale_artist(scale, action=action)
        return h

### group/ungroup (group is defined in FigObj)
    def ungroup(self):
        '''
        ungroup object
        if self does not have parent, it can not
        be ungrouped
        '''
        p = self.get_parent()
        if p is None:
            return
        ret = [child for name, child in self.get_children()]
        for child in ret:
            child.move(p, keep_zorder=True)
        self.destroy()
        return ret

    def set_parent(self, parent):
        super(FigGrp, self).set_parent(parent)
        for name, child in self.get_children():
            child.set_parent(self)

    def get_xrange(self, xrange=[None, None], scale='linear'):
        for name, child in self.get_children():
            xrange = child.get_xrange(xrange=xrange, scale=scale)
        return xrange

    def get_yrange(self, yrange=[None, None],
                   xrange=[None, None], scale='linear'):
        for name, child in self.get_children():
            yrange = child.get_yrange(xrange=xrange,
                                      yrange=yrange, scale=scale)
        return yrange

    def get_crange(self, crange=[None, None],
                   xrange=[None, None],
                   yrange=[None, None], scale='linear'):
        for name, child in self.get_children():
            crange = child.get_crange(crange=crange,
                                      xrange=xrange,
                                      yrange=yrange, scale=scale)
        return crange

    def get_zrange(self, zrange=[None, None],
                   xrange=[None, None],
                   yrange=[None, None], scale='linear'):
        for name, child in self.get_children():
            zrange = child.get_zrange(zrange=zrange,
                                      xrange=xrange,
                                      yrange=yrange, scale=scale)
        return zrange

    def set_container_idx(self, value, a=None):
        FigObj.set_container_idx(self, value, a=a)
        for name, child in self.get_children():
            child.set_container_idx(value, a=a)

    def set_caxis_idx(self, value, a):
        for name, child in self.get_children():
            child.set_caxis_idx(value, a=None)

    def get_axis_param_idx(self):
        for name, child in self.get_children():
            idx = child.get_axis_param_idx()
            if idx[0] is not None:
                return idx
        return [None]*4

    def set_zorder(self, z1, a=None):
        self.setp('zorder', z1)
        for a in self._artists:
            a.set_zorder(z1)
        for name, child in self.get_children():
            child.set_zorder(z1, a=None)
        self.set_bmp_update(False)

    def get_zorder(self, a=None):
        return self.getp('zorder')



from __future__ import print_function
#
#  example program to connect proxy_efit
#
import sys
import Pyro4
import Pyro4.util

sys.excepthook = Pyro4.util.excepthook

nshost = 'cmodws30.psfc.mit.edu'
nameserver = Pyro4.locateNS(host=nshost)
uri = nameserver.lookup("proxy_nete")
proxy = Pyro4.Proxy(uri)
print(proxy.run_idl(shot=1120710012, time=1.))



from __future__ import print_function
import wx
from ifigure.utils.cbook import parseStr
from ifigure.mto.py_code import PyData
from ifigure.mto.py_code import PyCode
from ifigure.mto.py_module import PyModule
from ifigure.mto.py_contents import NETCDFfile
import ifigure.events

import os
import re
import string
import sys
from numpy import *
from collections import OrderedDict
from netCDF4 import Dataset
######################################################
#         Setting for module file for py_module
#
#   General rule:
#      This file will be automatically loaded when
#      py_module object is created. Also, py_module
#      keeps track the file modification time. If
#      the file is updated, it will be automaticaaly
#      reloaded.


#      Strong recommendation : make module "independent".

#      Py_Modules does not check the dependency of
#      modules.
#      If moduels used in Py_Modules depends on
#      each other by for example module variable,
#      it will cause complicate  module loading
#      order-dependency problem.
#
#   name of module
module_name = 'netCDF'
class_name = 'netCDF'
#   module_evt_handler
#   functions which can be called from project tree
#
#    (menu name, function name, a flat to call skip()
#     after running function)
#
#   By default these function should return None
#   or True
#   if it return False, ifigure stops exectuion at
#   this module
#
menu = [("Import...",   "onLoadFile", True),
        ("Export...",   "onExportFile", False),
        ("Update Tree", "onUpdateTree", True),
        ("Update File", "onUpdateFile", True),        
        ("Check NC Format", "onCheckFormat", True),
        ("Write NC File", "onWriteNCFile", True), ]
method = ['onLoadFile', 'onUpdateTree', 'onUpdateFile',
          'onExportFile', 'onWriteNCFile', 'onCheckFormat',
          'init', 'init_after_load']
icon = 'data.png'
can_have_child = False
has_private_owndir = False
nosave_var0 = True
######################################################

modename = 'nc4_pathmode'
pathname = 'nc4_path'
extname = 'nc4_ext'


def init(self, *args, **kargs):
    #   a function called when py_module is initialized
    self.td.mk_owndir()
    if 'src' not in kargs:
        self.onLoadFile()
    # self.onLoadFile()


def do_load_netcdf_file(nm, g, keylist=None):

    if keylist is None:
        keylist = []

    nm["dimensions"] = NETCDFfile()
    nm["dimensions"].nc_path = keylist+["dimensions"]
    
    for dimname in g.dimensions:
        nm["dimensions"].var[dimname] = len(g.dimensions[dimname])
    for attname in g.ncattrs():
        nm.var[attname] = getattr(g, attname)

    nm["variables"] = NETCDFfile()
    nm["variables"].nc_path = keylist+["variables"]

    for varname in g.variables:
        nm["variables"][varname] = NETCDFfile()
        nm["variables"][varname].nc_path = keylist+["variables", varname]

        src = g.variables[varname]
        var = OrderedDict()
        var["dimensions"] = src.dimensions
        var["dtype"] = src.dtype
        var["ndim"] = src.ndim
        var["shape"] = src.shape
        for attname in src.ncattrs():
            var[attname] = getattr(src, attname)
        nm["variables"][varname].var = var
    # walk thorough all sub groups

    i = 1
    
    for sub_g in g.groups:
        #name = 'group'+str(i)
        name = sub_g
        #nm["variables"][name] = NETCDFfile()
        #nm["variables"][name].nc_path = keylist+["variables", name]
        #do_load_netcdf_file(nm["variables"][name], g[sub_g],
        #                    keylist=keylist+["variables", name])
        nm[name] = NETCDFfile()
        nm[name].nc_path = keylist+[name,]
        do_load_netcdf_file(nm[name], g[sub_g],
                            keylist=keylist+[name])
        i = i+1


def write_file(td, filename='/home/shiraiwa/test.nc',
               ignore_changes=False, format='NETCDF3_CLASSIC'):
    import numpy as np
    nm = td.getvar0()
    dtype2str = {np.dtype('float64'):  'f8',
                 np.dtype('float32'):  'f4',
                 np.dtype('int64'): 'i8',
                 np.dtype('int32'): 'i4',
                 np.dtype('int16'): 'i2',
                 np.dtype('int8'):  'i1',
                 np.dtype('uint64'): 'u8',
                 np.dtype('uint32'): 'u4',
                 np.dtype('uint16'): 'u2',
                 np.dtype('uint8'):  'u1',
                 np.dtype('S1'):  'S1', }

    rootgrp = Dataset(filename, 'w', format=format)
    for key in nm.var:
        setattr(rootgrp, key, nm.var[key])
    for key in nm['dimensions'].var:
        rootgrp.createDimension(key, nm['dimensions'].var[key])
    for key in nm["variables"]:
        dim = nm["variables"][key].var['dimensions']
        datatype = dtype2str[nm["variables"][key].var['dtype']]
        variable = rootgrp.createVariable(key, datatype, dim)
        for attr in nm["variables"][key].var:
            if not attr in ["dimensions", "dtype", "ndim", "shape"]:
                setattr(variable, attr, nm["variables"][key].var[attr])
        if ignore_changes or not nm["variables"][key]._data_loaded:
            variable[:] = nm["variables"][key].nc_eval(td)
        else:
            variable[:] = nm["variables"][key]._data

    rootgrp.close()
    print('write NC file completed. ' + filename)


def load_file(fname, check_format=False):
    import subprocess
    import shlex

    # read file
    compress = False
    if fname[-3:] == '.gz':
        command = 'gunzip '+fname
#       print command
        p = subprocess.Popen(shlex.split(command))
        p.wait()
        compress = True
        fname2 = fname[:-3]
#       print fname2
        if not os.path.exists(fname2):
            print(('!!!!!!!!!!!! netcdf file is not found', fname2))
            return
        g = Dataset(fname2, 'r')
    else:
        #       print fname
        if not os.path.exists(fname):
            print(('!!!!!!!!!!!! netcdf file is not found', fname))
            return
        g = Dataset(fname, 'r')
    if check_format:
        f = g.data_model
        g.close()
        return f
    nm = NETCDFfile()
    
    do_load_netcdf_file(nm, g)
    g.close()
    if compress:
        command = 'gzip '+fname2
        print(command)
        subprocess.Popen(shlex.split(command))
    return nm
#    obj.setvar0(nm)


def load_netcdf_file(obj):
    file = obj.path2fullpath(modename=modename,
                             pathname=pathname)

    if file == '':
        return
    # clean first
    for name, child in obj.get_children():
        child.destroy()

    nm = load_file(file)
    if nm is None:
        obj.setvar0(None)
        return
    obj.setvar0(nm)


def onWriteNCFile(self, e=None):
    open_dlg = wx.FileDialog(wx.GetApp().TopWindow,
                             message='Enter New NC file',
                             style=wx.FD_SAVE,
                             wildcard='NETCDF3_CLASSIC|*.nc|NETCDF3_64Bit|*.nc|NETCDF4(HDF5 with CDF4API)|*.hdf|NETCDF4_CLASIC(only CDF3)|*.hdf')
    if open_dlg.ShowModal() != wx.ID_OK:
        open_dlg.Destroy()
        return
    else:
        filename = str(open_dlg.GetPath())
        idx = open_dlg.GetFilterIndex()
        open_dlg.Destroy()
    format = ('NETCDF3_CLASSIC', 'NETCDF3_64BIT',
              'NETCDF4', 'NETCDF4_CLASSIC')[idx]
    write_file(self.td, filename=filename, format=format)


def onCheckFormat(self, e=None):
    obj = self.td
    file = obj.path2fullpath(modename=modename,
                             pathname=pathname)
    if file == '':
        return
    format = load_file(file, check_format=True)
    print(format)


def onUpdateTree(self, e=None):
    obj = self.td
    load_netcdf_file(obj)
    
def onUpdateFile(self, e=None):
    obj = self.td
    file = obj.path2fullpath(modename=modename,
                             pathname=pathname)
    if file == '':
        return
    
    file2 = os.path.join(os.path.dirname(file),
                         '_tmp_'+os.path.basename(file))
    format = load_file(file, check_format=True)
    write_file(self.td, filename=file2, format=format)
    os.rename(file2, file)
    

def onLoadFile(self, e=None, file=''):
    from ifigure.utils.addon_utils import onLoadFile

    if file != '':
        self.td.set_path_pathmode(file, modename, pathname, extname)
        ret = True
    else:
        ret = onLoadFile(self.td, message="Select NETCDF4 File",
                         modename=modename,
                         pathname=pathname,
                         extname=extname,
                         wildcard='nc(*.nc)|*.nc|nc.gz(*.nc.gz)|*.nc.gz|cdf(*.cdf)|*.cdf|Any|*',
                         )
    if ret:
        load_netcdf_file(self.td)
        ifigure.events.SendChangedEvent(self.td)


def onExportFile(self, evt=None):
    '''
    for now, it just copies the file
    '''
    obj = self.td
    file = obj.path2fullpath(modename=modename,
                             pathname=pathname)
    dir = os.path.expanduser('~')
    open_dlg = wx.FileDialog(wx.GetApp().TopWindow,
                             message='Enter file name to save',
                             defaultDir=dir,
                             defaultFile=os.path.basename(file),
                             wildcard='Any|*',
                             style=wx.FD_SAVE)
    if open_dlg.ShowModal() != wx.ID_OK:
        open_dlg.Destroy()
        return
    else:
        filename = str(open_dlg.GetPath())
        open_dlg.Destroy()
    import shutil
    shutil.copyfile(file, filename)


def init_after_load(self, olist, nlist):
    obj = self.td
    load_netcdf_file(obj)


def set_contents(self, *args):
    obj = self[:]
    for x in args[:-1]:
        obj = obj[x]
    obj._data_loaded = True
    obj._data = args[-1]



import numpy as np
from scipy.spatial import Delaunay
from matplotlib.tri import Triangulation

use_mpl_tri = False


def tri_args(x, y, _tri=None):
    #    use_mpl_tri = False
    if use_mpl_tri:
        if _tri is None:
            _tri = Triangulation(x, y)
        args = [x, y, _tri.triangles]
    else:
        if _tri is None:
            _tri = delaunay(x, y)

        args = [x.flatten(), y.flatten(), _tri]

    return args, _tri


def delaunay(x, y):
    '''
    tri = delaunay(x, y)
    '''
    if len(x.shape) == 1 and len(y.shape) == 1:
        _tri = Delaunay(np.hstack((x.reshape(-1, 1),
                                   y.reshape(-1, 1),))).simplices.copy()

    elif len(x.shape) == 2 and len(y.shape) == 2:
        xx = np.arange(x.shape[1])
        yy = np.arange(x.shape[0])
        XX, YY = np.meshgrid(xx, yy)
        _tri = Delaunay(np.hstack((XX.flatten().reshape(-1, 1),
                                  YY.flatten().reshape(-1, 1),))).simplices.copy()
    else:
        import warning
        warning.warn(
            "triangulation input data has ndim > 2 or x and y have differnt ndim. Data is flattened first")
        _tri = Delaunay(np.hstack((x.flatten().reshape(-1, 1),
                                   y.flatten().reshape(-1, 1),))).simplices.copy()

    return _tri


def mask_inside(x, y, tri, edge, mask=None, op=np.logical_or):
    '''
    omit tringles 
    edge = (:, 2)
    '''
    pmid = np.transpose(np.vstack((x[tri].mean(axis=1), y[tri].mean(axis=1))))

    from ifigure.utils.geom import path_contain

    mask2 = [path_contain(edge, xy, check=False) for xy in pmid]
    if mask is not None:
        mask2 = op(mask, mask2)
    return np.array(mask2)


def mask_outside(x, y, tri, edge, mask=None, op=np.logical_or):
    pmid = np.vstack((x[tri].mean(axis=1), y[tri].mean(axis=1)))

    mask2 = np.logical_not(mask_inside(x, y, tri, edge,
                                       mask=None, op=op))
    if mask is not None:
        mask2 = op(mask, mask2)
    return np.array(mask2)


def mask_skew(x, y, tri, th=5, mask=None, op=np.logical_or, inv=False):
    ddd = np.vstack((np.sqrt((x[tri][:, 0] - x[tri][:, 1])**2+(y[tri][:, 0]-y[tri][:, 1])**2), np.sqrt((x[tri][:, 0] - x[tri][:, 2])
                                                                                                       ** 2+(y[tri][:, 0]-y[tri][:, 2])**2), np.sqrt((x[tri][:, 1] - x[tri][:, 2])**2+(y[tri][:, 1]-y[tri][:, 2])**2)))

    sk = np.max(ddd, 0)/np.min(ddd, 0)
    if inv:
        mask2 = sk < th
    else:
        mask2 = sk > th
    if mask is not None:
        mask2 = op(mask, mask2)
    return np.array(mask2)


def get_area(x, y, tri):
    ddd = (x[tri][:, 0]-x[tri][:, 1])*(y[tri][:, 0]-y[tri][:, 2]) - \
        (x[tri][:, 0]-x[tri][:, 2])*(y[tri][:, 0]-y[tri][:, 1])
    return np.abs(ddd)


def mask_area(x, y, tri, th=5, mask=None, op=np.logical_or, inv=False):
    ddd = get_area(x, y, tri)
    if inv:
        mask2 = ddd < th
    else:
        mask2 = ddd > th
    if mask is not None:
        mask2 = op(mask, mask2)
    return np.array(mask2)
